import numpy as np
import warnings


def interpolate_linear(start, end, steps):
    """ Interpolate series between start and end in given number of steps - linear interpolation """
    return np.linspace(start, end, steps)


def interpolate_logscale(start, end, steps):
    """ Interpolate series between start and end in given number of steps - logscale interpolation """
    if start <= 0.0:
        warnings.warn("Start of logscale interpolation must be positive!")
        start = 1e-5

    return np.logspace(np.log10(float(start)), np.log10(float(end)), steps)


def interpolate_cosine_single(start, end, coefficient):
    """ Cosine interpolation """
    cos_out = (np.cos(np.pi*coefficient) + 1) / 2.0
    return end + (start - end) * cos_out


def interpolate_linear_single(start, end, coefficient):
    """ Cosine interpolation """
    return start + (end - start) * coefficient


def interpolate_logscale_single(start, end, coefficient):
    """ Cosine interpolation """
    return np.exp(np.log(start) + (np.log(end) - np.log(start)) * coefficient)


INTERP_DICT = {
    'linear': interpolate_linear,
    'logscale': interpolate_logscale
}


INTERP_SINGLE_DICT = {
    'linear': interpolate_linear_single,
    'logscale': interpolate_logscale_single,
    'cosine': interpolate_cosine_single
}


def interpolate_series(start, end, steps, how='linear'):
    """ Interpolate series between start and end in given number of steps """
    return INTERP_DICT[how](start, end, steps)


def interpolate_single(start, end, coefficient, how='linear'):
    """ Interpolate single value between start and end in given number of steps """
    return INTERP_SINGLE_DICT[how](start, end, coefficient)




class EvaluatorMeta(type):
    """ Metaclass for Evaluator - gathers all provider methods in a class attribute """
    def __new__(mcs, name, bases, attributes):
        providers = {}

        for name, attr in attributes.items():
            if callable(attr):
                proper_name = getattr(attr, '_vel_evaluator_provides', None)

                if proper_name is not None:
                    providers[proper_name] = attr

        attributes['_providers'] = providers

        return super().__new__(mcs, name, bases, attributes)


class Evaluator(metaclass=EvaluatorMeta):
    """
    Different models may have different outputs and approach evaluating environment differently.

    Evaluator is an object that abstracts over that, providing unified interface between algorithms
    which just need certain outputs from models and models that may provide them in different ways.

    I'll try to maintain here a dictionary of possible common values that can be requested from the evaluator.
    Rollouts should communicate using the same names

    - rollout:estimated_returns
        - Bootstrapped return (sum of discounted future rewards) estimated using returns and value estimates
    - rollout:estimated_values
        - Value estimates from the model that was used to generate the rollout
    - rollout:estimated_advantages
        - Advantage of a rollout (state, action) pair by the model that was used to generate the rollout
    - rollout:actions
        - Actions performed in a rollout
    - rollout:logprobs
        - Logarithm of probability for **all** actions of a policy used to perform rollout
        (defined only for finite action spaces)
    - rollout:action:logprobs
        - Logarithm of probability only for selected actions
    - rollout:dones
        - Whether given observation is last in a trajectory
    - rollout:dones
        - Raw rewards received from the environment in this learning process
    - rollout:final_estimated_values
        - Value estimates for observation after final observation in the rollout
    - rollout:observations
        - Observations of the rollout
    - rollout:observations_next
        - Next observations in the rollout
    - rollout:weights
        - Error weights of rollout samples

    - model:logprobs
        - Logarithm of probability of **all** actions in an environment as in current model policy
        (defined only for finite action spaces)
    - model:q
        - Action-value for **all** actions
        (defined only for finite action spaces)
    - model:q_next
        - Action-value for **all** actions from the 'next' state in the rollout
        (defined only for finite action spaces)
    - model:entropy
        - Policy entropy for selected states
    - model:action:q
        - Action-value for actions selected in the rollout
    - model:model_action:q
        - Action-value for actions that model would perform (Deterministic policy only)
    - model:actions
        - Actions that model would perform (Deterministic policy only)
    - model:action:logprobs
        - Logarithm of probability for performed actions
    - model:policy_params
        - Parametrizations of policy for each state
    - model:estimated_values
        - Value estimates for each state
    - model:estimated_values_next
        - Value estimates for 'next' state of each transition
    """

    @staticmethod
    def provides(name):
        """ Function decorator - value provided by the evaluator """
        def decorator(func):
            func._vel_evaluator_provides = name
            return func

        return decorator

    def __init__(self, rollout):
        self._storage = {}
        self.rollout = rollout

    def get(self, name):
        """
        Return a value from this evaluator.

        Because tensor calculated is cached, it may lead to suble bugs if the same value is used multiple times
        with and without no_grad() context.

        It is advised in such cases to not use no_grad and stick to .detach()
        """
        if name in self._storage:
            return self._storage[name]
        elif name in self._providers:
            value = self._storage[name] = self._providers[name](self)
            return value
        elif name.startswith('rollout:'):
            rollout_name = name[8:]
            value = self._storage[name] = self.rollout.batch_tensor(rollout_name)
            return value
        else:
            raise RuntimeError(f"Key {name} is not provided by this evaluator")

    def provide(self, name, value):
        """ Provide given value under specified name """
        self._storage[name] = value



import vel.util.intepolate as interpolate

from vel.api.base import Schedule


class LinearSchedule(Schedule):
    """ Interpolate variable linearly between start value and final value """

    def __init__(self, initial_value, final_value):
        self.initial_value = initial_value
        self.final_value = final_value

    def value(self, progress_indicator):
        """ Interpolate linearly between start and end """
        return interpolate.interpolate_linear_single(self.initial_value, self.final_value, progress_indicator)


def create(initial_value, final_value):
    """ Vel creation function """
    return LinearSchedule(initial_value, final_value)




import torch
import torch.nn as nn
import torch.nn.init as init

import gym.spaces as spaces


class DuelingQHead(nn.Module):
    """ Network head calculating Q-function value for each (discrete) action using two separate inputs. """
    def __init__(self, input_dim, action_space):
        super().__init__()

        # Q-function requires a discrete action space
        assert isinstance(action_space, spaces.Discrete)

        self.linear_layer_advantage = nn.Linear(input_dim, action_space.n)
        self.linear_layer_value = nn.Linear(input_dim, 1)
        self.action_space = action_space

    def reset_weights(self):
        for m in self.modules():
            if isinstance(m, nn.Linear):
                # init.kaiming_normal_(m.weight, mode='fan_out', nonlinearity='relu')
                init.orthogonal_(m.weight, gain=1.0)
                init.constant_(m.bias, 0.0)

    def forward(self, advantage_data, value_data):
        adv = self.linear_layer_advantage(advantage_data)
        value = self.linear_layer_value(value_data)
        # Advantage must be 0-centered

        return (adv - adv.mean(dim=1, keepdim=True)) + value

    def sample(self, q_values):
        """ Sample from epsilon-greedy strategy with given q-values """
        return q_values.argmax(dim=1)



"""
Code (partially) based on:
https://github.com/fastai/fastai/blob/master/fastai/layers.py
"""
import torch
import torch.nn as nn

from vel.util.tensor_util import one_hot_encoding


class AdaptiveConcatPool2d(nn.Module):
    """ Concat pooling - combined average pool and max pool """
    def __init__(self, sz=None):
        super().__init__()
        sz = sz or (1, 1)
        self.ap = nn.AdaptiveAvgPool2d(sz)
        self.mp = nn.AdaptiveMaxPool2d(sz)

    def forward(self, x):
        return torch.cat([self.mp(x), self.ap(x)], 1)


class Lambda(nn.Module):
    """ Simple torch lambda layer """
    def __init__(self, f):
        super().__init__()
        self.f = f

    def forward(self, x):
        return self.f(x)


class Flatten(nn.Module):
    """ Flatten input vector """
    def __init__(self):
        super().__init__()

    def forward(self, x):
        return x.view(x.size(0), -1)


class Identity(nn.Module):
    """ Identity transformation that doesn't do anything """
    def __init__(self):
        super().__init__()

    def forward(self, x):
        return x


class OneHotEncode(nn.Module):
    """ One-hot encoding layer """
    def __init__(self, num_classes):
        super().__init__()
        self.num_classes = num_classes

    def forward(self, x):
        return one_hot_encoding(x.view(-1), self.num_classes).view(list(x.shape) + [-1])




import torch
import torch.nn as nn
import torch.nn.init as init

from vel.api.base import RnnLinearBackboneModel, ModelFactory
from vel.rl.models.backbone.nature_cnn import NatureCnn


class NatureCnnLstmBackbone(RnnLinearBackboneModel):
    """
    Long-Short-Term Memory rnn cell together with DeepMind-style 'Nature' cnn preprocessing
    """

    def __init__(self, input_width: int, input_height: int, input_channels: int, cnn_output_dim: int=512,
                 hidden_units: int=128):
        super().__init__()

        self.hidden_units = hidden_units

        self.nature_cnn = NatureCnn(input_width, input_height, input_channels, cnn_output_dim)
        self.lstm = nn.LSTMCell(input_size=512, hidden_size=self.hidden_units)

    def reset_weights(self):
        """ Call proper initializers for the weights """
        self.nature_cnn.reset_weights()

        init.orthogonal_(self.lstm.weight_ih, gain=1.0)
        init.orthogonal_(self.lstm.weight_hh, gain=1.0)
        init.zeros_(self.lstm.bias_ih)
        init.zeros_(self.lstm.bias_hh)

    @property
    def output_dim(self) -> int:
        return self.hidden_units

    @property
    def state_dim(self) -> int:
        """ Initial state of the network """
        return 2 * self.hidden_units

    def forward(self, input_image, state):
        cnn_output = self.nature_cnn(input_image)

        hidden_state, cell_state = torch.split(state, self.hidden_units, 1)
        hidden_state, cell_state = self.lstm(cnn_output, (hidden_state, cell_state))

        new_state = torch.cat([hidden_state, cell_state], dim=1)

        return hidden_state, new_state


def create(input_width, input_height, input_channels=1, cnn_output_dim=512, hidden_units=128):
    def instantiate(**_):
        return NatureCnnLstmBackbone(
            input_width=input_width, input_height=input_height, input_channels=input_channels,
            cnn_output_dim=cnn_output_dim, hidden_units=hidden_units
        )

    return ModelFactory.generic(instantiate)


# Add this to make nicer scripting interface
NatureCnnFactory = create



import torch
import torch.nn.functional as F

from vel.api.metrics.averaging_metric import AveragingNamedMetric
from vel.rl.api import Trajectories
from vel.rl.api.base import OptimizerAlgoBase


def select_indices(tensor, indices):
    """ Select indices from tensor """
    return tensor.gather(1, indices.unsqueeze(1)).squeeze()


class AcerPolicyGradient(OptimizerAlgoBase):
    """ Actor-Critic with Experience Replay - policy gradient calculations """

    def __init__(self, model_factory, trust_region: bool=True, entropy_coefficient: float=0.01,
                 q_coefficient: float=0.5, rho_cap: float=10.0, retrace_rho_cap: float=1.0, max_grad_norm: float=None,
                 average_model_alpha=0.99, trust_region_delta=1.0):
        super().__init__(max_grad_norm)

        self.discount_factor = None

        self.trust_region = trust_region
        self.model_factory = model_factory

        self.entropy_coefficient = entropy_coefficient
        self.q_coefficient = q_coefficient

        self.rho_cap = rho_cap
        self.retrace_rho_cap = retrace_rho_cap

        # Trust region settings
        self.average_model = None
        self.average_model_initialized = False
        self.average_model_alpha = average_model_alpha
        self.trust_region_delta = trust_region_delta

    def initialize(self, settings, model, environment, device):
        """ Initialize policy gradient from reinforcer settings """
        self.discount_factor = settings.discount_factor

        if self.trust_region:
            self.average_model = self.model_factory.instantiate(action_space=environment.action_space).to(device)

    def update_average_model(self, model):
        """ Update weights of the average model with new model observation """
        if not self.average_model_initialized:
            # Initialize average model to have the same weights as the main model
            self.average_model.load_state_dict(model.state_dict())
            self.average_model_initialized = True
        else:
            for model_param, average_param in zip(model.parameters(), self.average_model.parameters()):
                # EWMA average model update
                average_param.data.mul_(self.average_model_alpha).add_(model_param.data * (1 - self.average_model_alpha))

    def calculate_gradient(self, batch_info, device, model, rollout):
        """ Calculate loss of the supplied rollout """
        assert isinstance(rollout, Trajectories), "ACER algorithm requires trajectory input"

        local_epsilon = 1e-6

        evaluator = model.evaluate(rollout)

        actions = evaluator.get('rollout:actions')
        rollout_probabilities = torch.exp(evaluator.get('rollout:logprobs'))

        # We calculate the trust-region update with respect to the average model
        if self.trust_region:
            self.update_average_model(model)

        logprobs = evaluator.get('model:logprobs')
        q = evaluator.get('model:q')

        # Selected action values
        action_logprobs = select_indices(logprobs, actions)
        action_q = select_indices(q, actions)

        # We only want to propagate gradients through specific variables
        with torch.no_grad():
            model_probabilities = torch.exp(logprobs)

            # Importance sampling correction - we must find the quotient of probabilities
            rho = model_probabilities / (rollout_probabilities + local_epsilon)

            # Probability quotient only for selected actions
            actions_rho = select_indices(rho, actions)

            # Calculate policy state values
            model_state_values = (model_probabilities * q).sum(dim=1)

            trajectory_rewards = rollout.transition_tensors['rewards']
            trajectory_dones = rollout.transition_tensors['dones']

            q_retraced = self.retrace(
                trajectory_rewards,
                trajectory_dones,
                action_q.reshape(trajectory_rewards.size()),
                model_state_values.reshape(trajectory_rewards.size()),
                actions_rho.reshape(trajectory_rewards.size()),
                rollout.rollout_tensors['final_estimated_values']
            ).flatten()

            advantages = q_retraced - model_state_values
            importance_sampling_coefficient = torch.min(actions_rho, self.rho_cap * torch.ones_like(actions_rho))

            explained_variance = 1 - torch.var(q_retraced - action_q) / torch.var(q_retraced)

        # Entropy of the policy distribution
        policy_entropy = torch.mean(model.entropy(logprobs))
        policy_gradient_loss = -torch.mean(advantages * importance_sampling_coefficient * action_logprobs)

        # Policy gradient bias correction
        with torch.no_grad():
            advantages_bias_correction = q - model_state_values.view(model_probabilities.size(0), 1)
            bias_correction_coefficient = F.relu(1.0 - self.rho_cap / (rho + local_epsilon))

        # This sum is an expectation with respect to action probabilities according to model policy
        policy_gradient_bias_correction_gain = torch.sum(
            logprobs * bias_correction_coefficient * advantages_bias_correction * model_probabilities,
            dim=1
        )

        policy_gradient_bias_correction_loss = - torch.mean(policy_gradient_bias_correction_gain)

        policy_loss = policy_gradient_loss + policy_gradient_bias_correction_loss

        q_function_loss = 0.5 * F.mse_loss(action_q, q_retraced)

        if self.trust_region:
            with torch.no_grad():
                average_evaluator = self.average_model.evaluate(rollout)
                average_action_logits = average_evaluator.get('model:logprobs')

            actor_loss = policy_loss - self.entropy_coefficient * policy_entropy
            q_loss = self.q_coefficient * q_function_loss

            actor_gradient = torch.autograd.grad(-actor_loss, logprobs, retain_graph=True)[0]

            # kl_divergence = model.kl_divergence(average_action_logits, action_logits).mean()
            # kl_divergence_grad = torch.autograd.grad(kl_divergence, action_logits, retain_graph=True)

            # Analytically calculated derivative of KL divergence on logits
            # That makes it hardcoded for discrete action spaces
            kl_divergence_grad_symbolic = - torch.exp(average_action_logits) / logprobs.size(0)

            k_dot_g = (actor_gradient * kl_divergence_grad_symbolic).sum(dim=-1)
            k_dot_k = (kl_divergence_grad_symbolic ** 2).sum(dim=-1)

            adjustment = (k_dot_g - self.trust_region_delta) / k_dot_k
            adjustment_clipped = adjustment.clamp(min=0.0)

            actor_gradient_updated = actor_gradient - adjustment_clipped.view(adjustment_clipped.size(0), 1)

            # Populate gradient from the newly updated fn
            logprobs.backward(gradient=-actor_gradient_updated, retain_graph=True)
            q_loss.backward(retain_graph=True)
        else:
            # Just populate gradient from the loss
            loss = policy_loss + self.q_coefficient * q_function_loss - self.entropy_coefficient * policy_entropy

            loss.backward()

        return {
            'policy_loss': policy_loss.item(),
            'policy_gradient_loss': policy_gradient_loss.item(),
            'policy_gradient_bias_correction': policy_gradient_bias_correction_loss.item(),
            'avg_q_selected': action_q.mean().item(),
            'avg_q_retraced': q_retraced.mean().item(),
            'q_loss': q_function_loss.item(),
            'policy_entropy': policy_entropy.item(),
            'advantage_norm': torch.norm(advantages).item(),
            'explained_variance': explained_variance.item(),
            'model_prob_std': model_probabilities.std().item(),
            'rollout_prob_std': rollout_probabilities.std().item()
        }

    def retrace(self, rewards, dones, q_values, state_values, rho, final_values):
        """ Calculate Q retraced targets """
        rho_bar = torch.min(torch.ones_like(rho) * self.retrace_rho_cap, rho)

        q_retraced_buffer = torch.zeros_like(rewards)

        next_value = final_values

        for i in reversed(range(rewards.size(0))):
            q_retraced = rewards[i] + self.discount_factor * next_value * (1.0 - dones[i])

            # Next iteration
            next_value = rho_bar[i] * (q_retraced - q_values[i]) + state_values[i]

            q_retraced_buffer[i] = q_retraced

        return q_retraced_buffer

    def metrics(self) -> list:
        """ List of metrics to track for this learning process """
        return [
            AveragingNamedMetric("q_loss"),
            AveragingNamedMetric("policy_entropy"),
            AveragingNamedMetric("policy_loss"),
            AveragingNamedMetric("policy_gradient_loss"),
            AveragingNamedMetric("policy_gradient_bias_correction"),
            AveragingNamedMetric("explained_variance"),
            AveragingNamedMetric("advantage_norm"),
            AveragingNamedMetric("grad_norm"),
            AveragingNamedMetric("model_prob_std"),
            AveragingNamedMetric("rollout_prob_std"),
            AveragingNamedMetric("avg_q_selected"),
            AveragingNamedMetric("avg_q_retraced")
        ]


def create(model, trust_region, entropy_coefficient, q_coefficient, max_grad_norm, rho_cap=10.0, retrace_rho_cap=1.0,
           average_model_alpha=0.99, trust_region_delta=1.0):
    return AcerPolicyGradient(
        trust_region=trust_region,
        model_factory=model,
        entropy_coefficient=entropy_coefficient,
        q_coefficient=q_coefficient,
        rho_cap=rho_cap,
        retrace_rho_cap=retrace_rho_cap,
        max_grad_norm=max_grad_norm,
        average_model_alpha=average_model_alpha,
        trust_region_delta=trust_region_delta
    )



import numpy as np

import nose.tools as t
import numpy.testing as nt

import torch
import torch.nn.functional as F
import torch.distributions as d

from vel.rl.modules.action_head import DiagGaussianActionHead, CategoricalActionHead


def test_sample_diag_gaussian():
    """ Test sampling from a multivariate gaussian distribution with a diagonal covariance matrix """
    head = DiagGaussianActionHead(1, 5)

    array = np.zeros((10000, 5, 2))

    sample = head.sample(torch.from_numpy(array))

    result_array = sample.detach().cpu().numpy()

    nt.assert_array_less(np.abs(result_array.mean(axis=0)), 0.1)
    nt.assert_array_less(result_array.std(axis=0), 1.1)
    nt.assert_array_less(0.9, result_array.std(axis=0))

    array2 = np.zeros((10000, 5, 2))
    array2[:, 0, 0] = 5.0
    array2[:, 0, 1] = np.log(10)

    sample2 = head.sample(torch.from_numpy(array2))

    result_array2 = sample2.detach().cpu().numpy()
    nt.assert_array_less(result_array2.mean(axis=0), np.array([5.3, 0.1, 0.1, 0.1, 0.1]))
    nt.assert_array_less(np.array([4.7, -0.1, -0.1, -0.1, -0.1]), result_array2.mean(axis=0))


def test_neglogp_diag_gaussian():
    """
    Test negative logarithm of likelihood of a multivariate gaussian distribution with a diagonal covariance matrix
    """
    head = DiagGaussianActionHead(1, 5)
    distrib = d.MultivariateNormal(torch.tensor([1.0, -1.0]), covariance_matrix=torch.tensor([[2.0, 0.0], [0.0, 0.5]]))

    pd_params = torch.tensor([[1.0, -1.0], [np.log(np.sqrt(2.0)), np.log(np.sqrt(0.5))]]).t()
    sample = head.sample(pd_params[None])

    log_prob1 = distrib.log_prob(sample)
    log_prob2 = head.logprob(sample, pd_params[None])

    nt.assert_allclose(log_prob1.detach().cpu().numpy(), log_prob2.detach().cpu().numpy(), rtol=1e-5)


def test_entropy_diag_gaussian():
    """
    Test entropy of a multivariate gaussian distribution with a diagonal covariance matrix
    """
    head = DiagGaussianActionHead(1, 5)
    distrib = d.MultivariateNormal(torch.tensor([1.0, -1.0]), covariance_matrix=torch.tensor([[2.0, 0.0], [0.0, 0.5]]))

    pd_params = torch.tensor([[1.0, -1.0], [np.log(np.sqrt(2.0)), np.log(np.sqrt(0.5))]]).t()

    entropy1 = distrib.entropy()
    entropy2 = head.entropy(pd_params[None])

    nt.assert_allclose(entropy1.detach().cpu().numpy(), entropy2.detach().cpu().numpy())


def test_kl_divergence_diag_gaussian():
    """
    Test kl divergence between multivariate gaussian distributions with a diagonal covariance matrix
    """
    head = DiagGaussianActionHead(1, 5)

    distrib1 = d.MultivariateNormal(torch.tensor([1.0, -1.0]), covariance_matrix=torch.tensor([[2.0, 0.0], [0.0, 0.5]]))
    distrib2 = d.MultivariateNormal(torch.tensor([0.3, 0.7]), covariance_matrix=torch.tensor([[1.8, 0.0], [0.0, 5.5]]))

    pd_params1 = torch.tensor([[1.0, -1.0], [np.log(np.sqrt(2.0)), np.log(np.sqrt(0.5))]]).t()
    pd_params2 = torch.tensor([[0.3, 0.7], [np.log(np.sqrt(1.8)), np.log(np.sqrt(5.5))]]).t()

    kl_div_1 = d.kl_divergence(distrib1, distrib2)
    kl_div_2 = head.kl_divergence(pd_params1[None], pd_params2[None])

    t.assert_almost_equal(kl_div_1.item(), kl_div_2.item(), places=3)


def test_sample_categorical():
    """
    Test sampling from a categorical distribution
    """
    head = CategoricalActionHead(1, 5)

    array = np.zeros((10000, 5))

    sample = head.sample(torch.from_numpy(array))

    result_array = sample.detach().cpu().numpy()

    nt.assert_array_less(np.abs(result_array.mean(axis=0)), 2.1)
    nt.assert_array_less(1.9, np.abs(result_array.mean(axis=0)))

    array2 = np.zeros((10000, 5))
    array2[:, 0:4] = -10.0
    array2[:, 4] = 10.0

    sample2 = head.sample(F.log_softmax(torch.from_numpy(array2), dim=1))
    result_array2 = sample2.detach().cpu().numpy()

    nt.assert_array_less(np.abs(result_array2.mean(axis=0)), 4.1)
    nt.assert_array_less(3.9, np.abs(result_array2.mean(axis=0)))


def test_neglogp_categorical():
    """
    Test negative logarithm of likelihood of a categorical distribution
    """
    head = CategoricalActionHead(1, 5)

    logits = F.log_softmax(torch.tensor([0.0, 1.0, 2.0, 3.0, 4.0]), dim=0)

    distrib = d.Categorical(logits=logits)

    actions = torch.tensor([0, 1, 2, 3, 4])

    log_p_1 = distrib.log_prob(actions)
    log_p_2 = head.logprob(actions, torch.stack([logits, logits, logits, logits, logits], dim=0))

    nt.assert_allclose(log_p_1.detach().cpu().numpy(), log_p_2.detach().cpu().numpy(), rtol=1e-5)


def test_entropy_categorical():
    """
    Test entropy of a categorical distribution
    """
    head = CategoricalActionHead(1, 5)

    logits = F.log_softmax(torch.tensor([0.0, 1.0, 2.0, 3.0, 4.0]), dim=0)

    distrib = d.Categorical(logits=logits)

    entropy1 = distrib.entropy()
    entropy2 = head.entropy(logits[None])

    nt.assert_allclose(entropy1.item(), entropy2.item())


def test_kl_divergence_categorical():
    """
    Test KL divergence between categorical distributions
    """
    head = CategoricalActionHead(1, 5)

    logits1 = F.log_softmax(torch.tensor([0.0, 1.0, 2.0, 3.0, 4.0]), dim=0)
    logits2 = F.log_softmax(torch.tensor([-1.0, 0.2, 5.0, 2.0, 8.0]), dim=0)

    distrib1 = d.Categorical(logits=logits1)
    distrib2 = d.Categorical(logits=logits2)

    kl_div_1 = d.kl_divergence(distrib1, distrib2)
    kl_div_2 = head.kl_divergence(logits1[None], logits2[None])

    nt.assert_allclose(kl_div_1.item(), kl_div_2.item(), rtol=1e-5)



from .base_metric import BaseMetric


class ValueMetric(BaseMetric):
    """ Base class for metrics that don't have state and just calculate a simple value """

    def __init__(self, name):
        super().__init__(name)

        self._metric_value = None

    def calculate(self, batch_info):
        """ Calculate value of a metric based on supplied data """
        self._metric_value = self._value_function(batch_info)

    def reset(self):
        """ Reset value of a metric """
        pass

    def value(self):
        """ Return current value for the metric """
        return self._metric_value

    def _value_function(self, batch_info):
        raise NotImplementedError




import os.path

import torchvision.datasets as ds

from vel.api.base import TrainingData


class ImageDirSource(ds.ImageFolder):
    pass


def create(model_config, path, num_workers, batch_size, augmentations=None, tta=None):
    """ Create an ImageDirSource with supplied arguments """
    if not os.path.isabs(path):
        path = model_config.project_top_dir(path)

    train_path = os.path.join(path, 'train')
    valid_path = os.path.join(path, 'valid')

    train_ds = ImageDirSource(train_path)
    val_ds = ImageDirSource(valid_path)

    return TrainingData(
        train_ds,
        val_ds,
        num_workers=num_workers,
        batch_size=batch_size,
        augmentations=augmentations,
        # test_time_augmentation=tta
    )



import time
import functools
import numpy as np

from collections import deque
from vel.openai.baselines import logger
from vel.util.random import set_seed

from .runner import Runner
from .buffer import Buffer
from .acer_model_proxy import AcerModelProxy

# from baselines import logger
# from baselines.common import set_global_seeds
# from baselines.common.policies import build_policy
# from baselines.common.tf_util import get_session, save_variables
#
# from baselines.a2c.utils import batch_to_seq, seq_to_batch
# from baselines.a2c.utils import cat_entropy_softmax
# from baselines.a2c.utils import Scheduler, find_trainable_variables
# from baselines.a2c.utils import EpisodeStats
# from baselines.a2c.utils import get_by_index, check_shape, avg_norm, gradient_add, q_explained_variance
# from baselines.acer.buffer import Buffer
# from baselines.acer.runner import Runner


def safemean(xs):
    return np.nan if len(xs) == 0 else np.mean(xs)


class Acer():
    def __init__(self, runner, model, buffer, log_interval):
        self.runner = runner
        self.model = model
        self.buffer = buffer
        self.log_interval = log_interval
        self.tstart = None
        # self.episode_stats = EpisodeStats(runner.nsteps, runner.nenv)
        self.epinfobuf = deque(maxlen=100)
        self.steps = None

    def call(self, on_policy):
        runner, model, buffer, steps = self.runner, self.model, self.buffer, self.steps

        if on_policy:
            enc_obs, obs, actions, rewards, mus, dones, masks, epinfos, rawobs = runner.run()

            self.epinfobuf.extend(epinfos)

            # self.episode_stats.feed(rewards, dones)

            if buffer is not None:
                buffer.put(enc_obs, actions, rewards, mus, dones, masks, rawobs)
        else:
            # get obs, actions, rewards, mus, dones from buffer.
            obs, actions, rewards, mus, dones, masks = buffer.get()

        # reshape stuff correctly
        obs = obs.reshape(runner.batch_ob_shape)
        actions = actions.reshape([runner.nbatch])
        rewards = rewards.reshape([runner.nbatch])
        mus = mus.reshape([runner.nbatch, runner.nact])
        dones = dones.reshape([runner.nbatch])
        masks = masks.reshape([runner.batch_ob_shape[0]])

        names_ops, values_ops = model.train(
            obs, actions, rewards, dones, mus, model.initial_state, masks, steps, on_policy=on_policy
        )

        if on_policy and (int(steps/runner.nbatch) % self.log_interval == 0):
            logger.record_tabular("total_timesteps", steps)
            logger.record_tabular("fps", int(steps/(time.time() - self.tstart)))
            # IMP: In EpisodicLife env, during training, we get done=True at each loss of life, not just at the terminal state.
            # Thus, this is mean until end of life, not end of episode.
            # For true episode rewards, see the monitor files in the log folder.

            # logger.record_tabular("mean_episode_length", self.episode_stats.mean_length())
            # logger.record_tabular("mean_episode_reward", self.episode_stats.mean_reward())

            logger.record_tabular('mean_episode_length', safemean([epinfo['l'] for epinfo in self.epinfobuf]))
            logger.record_tabular('mean_episode_reward', safemean([epinfo['r'] for epinfo in self.epinfobuf]))

            for name, val in zip(names_ops, values_ops):
                logger.record_tabular(name, float(val))
            logger.dump_tabular()


def learn(env, seed=None, nsteps=20, nstack=4, total_timesteps=int(80e6), q_coef=0.5, ent_coef=0.01,
          max_grad_norm=10, lr=7e-4, lrschedule='linear', rprop_epsilon=1e-5, rprop_alpha=0.99, gamma=0.99,
          log_interval=100, buffer_size=50000, replay_ratio=4, replay_start=10000, c=10.0,
          trust_region=True, alpha=0.99, delta=1, load_path=None, **network_kwargs):

    '''
    Main entrypoint for ACER (Actor-Critic with Experience Replay) algorithm (https://arxiv.org/pdf/1611.01224.pdf)
    Train an agent with given network architecture on a given environment using ACER.

    Parameters:
    ----------

    network:            policy network architecture. Either string (mlp, lstm, lnlstm, cnn_lstm, cnn, cnn_small, conv_only - see baselines.common/models.py for full list)
                        specifying the standard network architecture, or a function that takes tensorflow tensor as input and returns
                        tuple (output_tensor, extra_feed) where output tensor is the last network layer output, extra_feed is None for feed-forward
                        neural nets, and extra_feed is a dictionary describing how to feed state into the network for recurrent neural nets.
                        See baselines.common/policies.py/lstm for more details on using recurrent nets in policies

    env:                environment. Needs to be vectorized for parallel environment simulation.
                        The environments produced by gym.make can be wrapped using baselines.common.vec_env.DummyVecEnv class.

    nsteps:             int, number of steps of the vectorized environment per update (i.e. batch size is nsteps * nenv where
                        nenv is number of environment copies simulated in parallel) (default: 20)

    nstack:             int, size of the frame stack, i.e. number of the frames passed to the step model. Frames are stacked along channel dimension
                        (last image dimension) (default: 4)

    total_timesteps:    int, number of timesteps (i.e. number of actions taken in the environment) (default: 80M)

    q_coef:             float, value function loss coefficient in the optimization objective (analog of vf_coef for other actor-critic methods)

    ent_coef:           float, policy entropy coefficient in the optimization objective (default: 0.01)

    max_grad_norm:      float, gradient norm clipping coefficient. If set to None, no clipping. (default: 10),

    lr:                 float, learning rate for RMSProp (current implementation has RMSProp hardcoded in) (default: 7e-4)

    lrschedule:         schedule of learning rate. Can be 'linear', 'constant', or a function [0..1] -> [0..1] that takes fraction of the training progress as input and
                        returns fraction of the learning rate (specified as lr) as output

    rprop_epsilon:      float, RMSProp epsilon (stabilizes square root computation in denominator of RMSProp update) (default: 1e-5)

    rprop_alpha:        float, RMSProp decay parameter (default: 0.99)

    gamma:              float, reward discounting factor (default: 0.99)

    log_interval:       int, number of updates between logging events (default: 100)

    buffer_size:        int, size of the replay buffer (default: 50k)

    replay_ratio:       int, now many (on average) batches of data to sample from the replay buffer take after batch from the environment (default: 4)

    replay_start:       int, the sampling from the replay buffer does not start until replay buffer has at least that many samples (default: 10k)

    c:                  float, importance weight clipping factor (default: 10)

    trust_region        bool, whether or not algorithms estimates the gradient KL divergence between the old and updated policy and uses it to determine step size  (default: True)

    delta:              float, max KL divergence between the old policy and updated policy (default: 1)

    alpha:              float, momentum factor in the Polyak (exponential moving average) averaging of the model parameters (default: 0.99)

    load_path:          str, path to load the model from (default: None)

    **network_kwargs:               keyword arguments to the policy / network builder. See baselines.common/policies.py/build_policy and arguments to a particular type of network
                                    For instance, 'mlp' network architecture has arguments num_hidden and num_layers.

    '''

    print("Running Acer Simple")
    import pprint
    pprint.pprint(locals())

    set_seed(seed)

    # policy = build_policy(env, network, estimate_q=True, **network_kwargs)

    nenvs = env.num_envs
    ob_space = env.observation_space
    ac_space = env.action_space
    num_procs = len(env.remotes) if hasattr(env, 'remotes') else 1# HACK

    model = AcerModelProxy(
        # policy=policy,
        ob_space=ob_space,
        ac_space=ac_space,
        nenvs=nenvs,
        nsteps=nsteps,
        nstack=nstack,
        num_procs=num_procs,
        ent_coef=ent_coef,
        q_coef=q_coef,
        gamma=gamma,
        max_grad_norm=max_grad_norm,
        lr=lr,
        rprop_alpha=rprop_alpha,
        rprop_epsilon=rprop_epsilon,
        total_timesteps=total_timesteps,
        lrschedule=lrschedule,
        c=c,
        trust_region=trust_region,
        alpha=alpha, delta=delta)

    runner = Runner(env=env, model=model, nsteps=nsteps, nstack=nstack)

    if replay_ratio > 0:
        buffer = Buffer(env=env, nsteps=nsteps, nstack=nstack, size=buffer_size)
    else:
        buffer = None

    nbatch = nenvs*nsteps
    acer = Acer(runner, model, buffer, log_interval)
    acer.tstart = time.time()

    for acer.steps in range(0, total_timesteps, nbatch): #nbatch samples, 1 on_policy call and multiple off-policy calls
        # if acer.steps >= 2000000:
        #     1+2

        acer.call(on_policy=True)

        if replay_ratio > 0 and buffer.has_atleast(replay_start):
            n = np.random.poisson(replay_ratio)
            for _ in range(n):
                acer.call(on_policy=False)  # no simulation steps in this

    return model




import torch.nn as nn
import torch.nn.init as init


class ValueHead(nn.Module):
    """ Network head for value determination """
    def __init__(self, input_dim):
        super().__init__()

        self.linear_layer = nn.Linear(input_dim, 1)

    def reset_weights(self):
        # init.kaiming_normal_(m.weight, mode='fan_out', nonlinearity='relu')
        init.orthogonal_(self.linear_layer.weight, gain=1.0)
        init.constant_(self.linear_layer.bias, 0.0)

    def forward(self, input_data):
        return self.linear_layer(input_data)[:, 0]



from gym import Env
from gym.envs.registration import EnvSpec
from vel.openai.baselines.common.vec_env import VecEnv


class EnvFactory:
    """ Base class for environment factory """

    def specification(self) -> EnvSpec:
        """ Return environment specification """
        raise NotImplementedError

    def instantiate(self, seed=0, serial_id=1, preset='default', extra_args=None) -> Env:
        """ Create a new Env instance """
        raise NotImplementedError


class VecEnvFactory:
    """ Base class for vector environment factory """

    def instantiate(self, parallel_envs, seed=0, preset='default') -> VecEnv:
        """ Create a new VecEnv instance """
        raise NotImplementedError

    def instantiate_single(self, seed=0, preset='default') -> VecEnv:
        """ Create a new VecEnv instance - single """
        raise NotImplementedError




import gym
import torch

from vel.api.base import LinearBackboneModel, Model, ModelFactory
from vel.rl.api import Rollout, Evaluator
from vel.rl.modules.action_head import ActionHead
from vel.rl.modules.q_head import QHead


class QPolicyGradientEvaluator(Evaluator):
    """ Evaluator for QPolicyGradientModel """
    def __init__(self, model: 'QPolicyGradientModel', rollout: Rollout):
        super().__init__(rollout)

        self.model = model

        observations = self.get('rollout:observations')
        logprobs, q = model(observations)

        self.provide('model:logprobs', logprobs)
        self.provide('model:q', q)

    @Evaluator.provides('model:action:logprobs')
    def model_action_logprobs(self):
        actions = self.get('rollout_actions')
        logprobs = self.get('model:logprobs')
        return self.model.action_head.logprob(actions, logprobs)


class QPolicyGradientModel(Model):
    """ Custom heads for a policy gradient model with a action-value head """

    def __init__(self, backbone: LinearBackboneModel, action_space: gym.Space):
        super().__init__()

        assert isinstance(action_space, gym.spaces.Discrete)

        self.backbone = backbone

        self.action_head = ActionHead(
            input_dim=self.backbone.output_dim,
            action_space=action_space
        )

        self.q_head = QHead(
            input_dim=self.backbone.output_dim,
            action_space=action_space
        )

    def reset_weights(self):
        """ Initialize properly model weights """
        self.backbone.reset_weights()
        self.action_head.reset_weights()
        self.q_head.reset_weights()

    def forward(self, observations):
        """ Calculate model outputs """
        base_output = self.backbone(observations)

        policy_params = self.action_head(base_output)
        q = self.q_head(base_output)

        return policy_params, q

    def step(self, observation, argmax_sampling=False):
        """ Select actions based on model's output """
        policy_params, q = self(observation)
        actions = self.action_head.sample(policy_params, argmax_sampling=argmax_sampling)

        # log probability - we can do that, because we support only discrete action spaces
        logprobs = self.action_head.logprob(actions, policy_params)

        return {
            'actions': actions,
            'values': q,
            'logprobs': policy_params,
            'action_logprobs': logprobs
        }

    def evaluate(self, rollout: Rollout) -> QPolicyGradientEvaluator:
        """ Evaluate model on a rollout """
        return QPolicyGradientEvaluator(self, rollout)

        # observations = rollout.batch_tensor('observations')
        # actions = rollout.batch_tensor('actions')
        #
        # policy_params, q = self(observations)
        # logprobs = self.action_head.logprob(actions, policy_params)
        #
        # return {
        #     'action_pd_params': policy_params,
        #     'actiologprobs': logprobs,
        #     'q': q
        # }

    def value(self, observation):
        """ Calculate only value head for given state """
        policy_params, q = self(observation)

        # Expectation of Q value with respect to action
        return (torch.exp(policy_params) * q).sum(dim=1)

    def entropy(self, action_logits):
        """ Entropy of a probability distribution """
        return self.action_head.entropy(action_logits)

    def kl_divergence(self, logits_q, logits_p):
        """ Calculate KL-divergence between two probability distributions """
        return self.action_head.kl_divergence(logits_q, logits_p)


class QPolicyGradientModelFactory(ModelFactory):
    """ Factory  class for policy gradient models """
    def __init__(self, backbone: ModelFactory):
        self.backbone = backbone

    def instantiate(self, **extra_args):
        """ Instantiate the model """
        backbone = self.backbone.instantiate(**extra_args)
        return QPolicyGradientModel(backbone, extra_args['action_space'])


def create(backbone: ModelFactory):
    return QPolicyGradientModelFactory(backbone=backbone)



import numpy as np

from vel.rl.buffers.deque_multi_env_buffer_backend import DequeMultiEnvBufferBackend


class Buffer(object):
    # gets obs, actions, rewards, mu's, (states, masks), dones
    def __init__(self, env, nsteps, nstack, size=50000):
        self.nenv = env.num_envs
        self.nsteps = nsteps
        self.nh, self.nw, self.nc = env.observation_space.shape
        self.nstack = nstack
        self.nbatch = self.nenv * self.nsteps
        self.size = size // (self.nsteps)  # Each loc contains nenv * nsteps frames, thus total buffer is nenv * size frames

        # Memory
        self.enc_obs = None
        self.actions = None
        self.rewards = None
        self.mus = None
        self.dones = None
        self.masks = None

        # Size indexes
        self.next_idx = 0
        self.num_in_buffer = 0

        self.vel_buffer = DequeMultiEnvBufferBackend(
            buffer_capacity=size,
            num_envs=self.nenv,
            observation_space=env.observation_space,
            action_space=env.action_space,
            extra_data={
                'action_logits': np.zeros(
                    (size, env.num_envs, env.action_space.n), dtype=np.float32
                )
            },
            frame_stack_compensation=True
        )

    def has_atleast(self, frames):
        # Frames per env, so total (nenv * frames) Frames needed
        # Each buffer loc has nenv * nsteps frames
        return self.num_in_buffer >= (frames // self.nsteps)

    def can_sample(self):
        return self.num_in_buffer > 0

    # Generate stacked frames
    def decode(self, enc_obs, dones):
        # enc_obs has shape [nenvs, nsteps + nstack, nh, nw, nc]
        # dones has shape [nenvs, nsteps, nh, nw, nc]
        # returns stacked obs of shape [nenv, (nsteps + 1), nh, nw, nstack*nc]
        nstack, nenv, nsteps, nh, nw, nc = self.nstack, self.nenv, self.nsteps, self.nh, self.nw, self.nc
        y = np.empty([nsteps + nstack - 1, nenv, 1, 1, 1], dtype=np.float32)
        obs = np.zeros([nstack, nsteps + nstack, nenv, nh, nw, nc], dtype=np.uint8)
        x = np.reshape(enc_obs, [nenv, nsteps + nstack, nh, nw, nc]).swapaxes(1,
                                                                              0)  # [nsteps + nstack, nenv, nh, nw, nc]
        y[3:] = np.reshape(1.0 - dones, [nenv, nsteps, 1, 1, 1]).swapaxes(1, 0)  # keep
        y[:3] = 1.0
        # y = np.reshape(1 - dones, [nenvs, nsteps, 1, 1, 1])
        for i in range(nstack):
            obs[-(i + 1), i:] = x
            # obs[:,i:,:,:,-(i+1),:] = x
            x = x[:-1] * y
            y = y[1:]
        return np.reshape(obs[:, 3:].transpose((2, 1, 3, 4, 0, 5)), [nenv, (nsteps + 1), nh, nw, nstack * nc])

    def put(self, enc_obs, actions, rewards, mus, dones, masks, rawobs):
        # enc_obs [nenv, (nsteps + nstack), nh, nw, nc]
        # actions, rewards, dones [nenv, nsteps]
        # mus [nenv, nsteps, nact]

        if self.enc_obs is None:
            self.enc_obs = np.empty([self.size] + list(enc_obs.shape), dtype=np.uint8)
            self.actions = np.empty([self.size] + list(actions.shape), dtype=np.int32)
            self.rewards = np.empty([self.size] + list(rewards.shape), dtype=np.float32)
            self.mus = np.empty([self.size] + list(mus.shape), dtype=np.float32)
            self.dones = np.empty([self.size] + list(dones.shape), dtype=np.bool)
            self.masks = np.empty([self.size] + list(masks.shape), dtype=np.bool)

        self.enc_obs[self.next_idx] = enc_obs
        self.actions[self.next_idx] = actions
        self.rewards[self.next_idx] = rewards
        self.mus[self.next_idx] = mus
        self.dones[self.next_idx] = dones
        self.masks[self.next_idx] = masks

        self.next_idx = (self.next_idx + 1) % self.size
        self.num_in_buffer = min(self.size, self.num_in_buffer + 1)

        # for i in range(self.nsteps):


        print("STOP")

    def take(self, x, idx, envx):
        nenv = self.nenv
        out = np.empty([nenv] + list(x.shape[2:]), dtype=x.dtype)
        for i in range(nenv):
            out[i] = x[idx[i], envx[i]]
        return out

    def get(self):
        # returns
        # obs [nenv, (nsteps + 1), nh, nw, nstack*nc]
        # actions, rewards, dones [nenv, nsteps]
        # mus [nenv, nsteps, nact]
        nenv = self.nenv
        assert self.can_sample()

        # Sample exactly one id per env. If you sample across envs, then higher correlation in samples from same env.
        idx = np.random.randint(0, self.num_in_buffer, nenv)
        envx = np.arange(nenv)

        take = lambda x: self.take(x, idx, envx)  # for i in range(nenv)], axis = 0)
        dones = take(self.dones)
        enc_obs = take(self.enc_obs)
        obs = self.decode(enc_obs, dones)
        actions = take(self.actions)
        rewards = take(self.rewards)
        mus = take(self.mus)
        masks = take(self.masks)

        print("STOP")

        return obs, actions, rewards, mus, dones, masks



from .algo_base import AlgoBase, OptimizerAlgoBase
from .env_base import EnvFactory, VecEnvFactory
from .env_roller import EnvRollerBase, ReplayEnvRollerBase, EnvRollerFactory, ReplayEnvRollerFactory
from .reinforcer_base import ReinforcerBase, ReinforcerFactory



import torch

import numbers

from vel.api.metrics.averaging_metric import AveragingNamedMetric
from vel.rl.api.base import OptimizerAlgoBase
from vel.math.functions import explained_variance
from vel.schedules.constant import ConstantSchedule


class PpoPolicyGradient(OptimizerAlgoBase):
    """ Proximal Policy Optimization - https://arxiv.org/abs/1707.06347 """
    def __init__(self, entropy_coefficient, value_coefficient, cliprange, max_grad_norm, normalize_advantage=True):
        super().__init__(max_grad_norm)

        self.entropy_coefficient = entropy_coefficient
        self.value_coefficient = value_coefficient
        self.normalize_advantage = normalize_advantage

        if isinstance(cliprange, numbers.Number):
            self.cliprange = ConstantSchedule(cliprange)
        else:
            self.cliprange = cliprange

    def calculate_gradient(self, batch_info, device, model, rollout):
        """ Calculate loss of the supplied rollout """
        evaluator = model.evaluate(rollout)

        # Part 0.0 - Rollout values
        advantages = evaluator.get('rollout:estimated_advantages')
        rollout_values = evaluator.get('rollout:estimated_values')
        rollout_action_logprobs = evaluator.get('rollout:action:logprobs')
        returns = evaluator.get('rollout:estimated_returns')

        # PART 0.1 - Model evaluation
        entropy = evaluator.get('model:entropy')
        model_values = evaluator.get('model:estimated_values')
        model_action_logprobs = evaluator.get('model:action:logprobs')

        # Select the cliprange
        current_cliprange = self.cliprange.value(batch_info['progress'])

        # Normalize the advantages?
        if self.normalize_advantage:
            advantages = (advantages - advantages.mean()) / (advantages.std() + 1e-8)

        # PART 1 - policy entropy
        policy_entropy = torch.mean(entropy)

        # PART 2 - value function
        value_output_clipped = rollout_values + torch.clamp(
            model_values - rollout_values, -current_cliprange, current_cliprange
        )
        value_loss_part1 = (model_values - returns).pow(2)
        value_loss_part2 = (value_output_clipped - returns).pow(2)
        value_loss = 0.5 * torch.mean(torch.max(value_loss_part1, value_loss_part2))

        # PART 3 - policy gradient loss
        ratio = torch.exp(model_action_logprobs - rollout_action_logprobs)

        pg_loss_part1 = -advantages * ratio
        pg_loss_part2 = -advantages * torch.clamp(ratio, 1.0 - current_cliprange, 1.0 + current_cliprange)
        policy_loss = torch.mean(torch.max(pg_loss_part1, pg_loss_part2))

        loss_value = (
            policy_loss - self.entropy_coefficient * policy_entropy + self.value_coefficient * value_loss
        )

        loss_value.backward()

        with torch.no_grad():
            approx_kl_divergence = 0.5 * torch.mean((model_action_logprobs - rollout_action_logprobs).pow(2))
            clip_fraction = torch.mean((torch.abs(ratio - 1.0) > current_cliprange).to(dtype=torch.float))

        return {
            'policy_loss': policy_loss.item(),
            'value_loss': value_loss.item(),
            'policy_entropy': policy_entropy.item(),
            'approx_kl_divergence': approx_kl_divergence.item(),
            'clip_fraction': clip_fraction.item(),
            'advantage_norm': torch.norm(advantages).item(),
            'explained_variance': explained_variance(returns, rollout_values)
        }

    def metrics(self) -> list:
        """ List of metrics to track for this learning process """
        return [
            AveragingNamedMetric("policy_loss"),
            AveragingNamedMetric("value_loss"),
            AveragingNamedMetric("policy_entropy"),
            AveragingNamedMetric("approx_kl_divergence"),
            AveragingNamedMetric("clip_fraction"),
            AveragingNamedMetric("grad_norm"),
            AveragingNamedMetric("advantage_norm"),
            AveragingNamedMetric("explained_variance")
        ]


def create(entropy_coefficient, value_coefficient, cliprange, max_grad_norm, normalize_advantage=True):
    return PpoPolicyGradient(
        entropy_coefficient, value_coefficient, cliprange, max_grad_norm, normalize_advantage=normalize_advantage
    )



import os
import yaml

from vel.exceptions import VelException


class Dummy:
    """ Dummy instance placeholder """
    pass


DUMMY_VALUE = Dummy()


class Variable:
    """ Project configuration variable to be popupated from command line """
    @classmethod
    def parameter_constructor(cls, loader, node):
        """ Construct variable instance from yaml node """
        value = loader.construct_scalar(node)

        if isinstance(value, str):
            if '=' in value:
                (varname, varvalue) = Parser.parse_equality(value)
                return cls(varname, varvalue)
            else:
                return cls(value)
        else:
            return cls(value)

    def __init__(self, name, default_value=DUMMY_VALUE):
        self.name = name
        self.default_value = default_value

    def __repr__(self):
        return f"{self.__class__.__name__}(name={self.name})"

    def resolve(self, parameters):
        """ Resolve given variable """
        raise NotImplementedError


class Parameter(Variable):
    """ Parameter that gets propagated from command line options """
    def resolve(self, parameters):
        """ Resolve given variable """
        if self.default_value == DUMMY_VALUE:
            if self.name in parameters:
                return parameters[self.name]
            else:
                raise VelException(f"Undefined parameter: {self.name}")
        else:
            return parameters.get(self.name, self.default_value)


class EnvironmentVariable(Variable):
    """ Parameter that gets propagated from environment """
    def resolve(self, _):
        """ Resolve given variable """
        if self.default_value == DUMMY_VALUE:
            if self.name in os.environ:
                return os.environ[self.name]
            else:
                raise VelException(f"Undefined environment variable: {self.name}")
        else:
            return os.environ.get(self.name, self.default_value)


class Parser:
    """ Parse configuration values """
    IS_LOADED = False

    @classmethod
    def register(cls):
        """ Register variable handling in YAML """
        if not cls.IS_LOADED:
            cls.IS_LOADED = True

            yaml.add_constructor('!param', Parameter.parameter_constructor, Loader=yaml.SafeLoader)
            yaml.add_constructor('!env', EnvironmentVariable.parameter_constructor, Loader=yaml.SafeLoader)

    @classmethod
    def parse(cls, stream):
        """ Parse the stream into a Python object """
        cls.register()
        return yaml.safe_load(stream)

    @classmethod
    def parse_equality(cls, equality_string):
        """ Parse some simple equality statements """
        cls.register()
        assert '=' in equality_string, "There must be an '=' sign in the equality"
        [left_side, right_side] = equality_string.split('=', 1)

        left_side_value = yaml.safe_load(left_side.strip())
        right_side_value = yaml.safe_load(right_side.strip())

        assert isinstance(left_side_value, str), "Left side of equality must be a string"

        return left_side_value, right_side_value



from .callback import Callback
from .model import Model, BackboneModel, LinearBackboneModel, SupervisedModel, RnnLinearBackboneModel, RnnModel
from .model_factory import ModelFactory
from .optimizer import OptimizerFactory
from .schedule import Schedule
from .scheduler import SchedulerFactory
from .source import Source, TrainingData
from .storage import Storage
from .train_phase import TrainPhase, EmptyTrainPhase



"""
Code based on
https://github.com/fastai/fastai/blob/master/fastai/models/cifar10/resnext.py
"""
import torch.nn as nn
import torch.nn.functional as F


class ResNeXtBottleneck(nn.Module):
    """
    RexNeXt bottleneck type C (https://github.com/facebookresearch/ResNeXt/blob/master/models/resnext.lua)
    """

    def __init__(self, in_channels, out_channels, cardinality, divisor, stride=1):
        super(ResNeXtBottleneck, self).__init__()

        self.cardinality = cardinality
        self.stride = stride
        self.divisor = divisor

        # D is a size of a single group
        # Intermediate layers have D * C channels
        D = out_channels // divisor
        C = cardinality

        if stride != 1 or in_channels != out_channels:
            self.shortcut = nn.Sequential(
                nn.Conv2d(in_channels, out_channels, kernel_size=1, stride=stride, bias=False),
                nn.BatchNorm2d(out_channels),
            )
        else:
            self.shortcut = None

        self.conv_reduce = nn.Conv2d(in_channels, D * C, kernel_size=1, stride=1, padding=0, bias=False)
        self.bn_reduce = nn.BatchNorm2d(D*C)

        self.conv_conv = nn.Conv2d(D*C, D*C, kernel_size=3, stride=stride, padding=1, groups=cardinality, bias=False)
        self.bn = nn.BatchNorm2d(D*C)

        self.conv_expand = nn.Conv2d(D * C, out_channels, kernel_size=1, stride=1, padding=0, bias=False)
        self.bn_expand = nn.BatchNorm2d(out_channels)


    def forward(self, x):
        bottleneck = self.conv_reduce(x)
        bottleneck = F.relu(self.bn_reduce(bottleneck), inplace=True)

        bottleneck = self.conv_conv(bottleneck)
        bottleneck = F.relu(self.bn(bottleneck), inplace=True)

        bottleneck = self.conv_expand(bottleneck)
        bottleneck = self.bn_expand(bottleneck)

        if self.shortcut is not None:
            residual = self.shortcut(x)
        else:
            residual = x

        return F.relu(residual + bottleneck, inplace=True)



from .callback import Callback


class SchedulerFactory:
    """ Factory class for various schedulers """

    def instantiate(self, optimizer, last_epoch=-1) -> Callback:
        raise NotImplementedError



import torch
import numpy as np
import bisect
import typing

from vel.api import Learner, TrainingInfo, ModelConfig
from vel.api.base import TrainPhase


class PhaseTrainCommand:
    """ Training  command - learn according to a set of phases """

    def __init__(self, model_config: ModelConfig, model_factory, source, storage, phases: typing.List[TrainPhase],
                 callbacks=None, restart=True):
        self.model_config = model_config
        self.model_factory = model_factory
        self.source = source
        self.storage = storage
        self.phases = phases
        self.ladder = self._build_phase_ladder(phases)
        self.full_number_of_epochs = sum(p.number_of_epochs for p in phases)
        self.callbacks = callbacks if callbacks is not None else []
        self.restart = restart

    @staticmethod
    def _build_phase_ladder(phases):
        """ Build a ladder of learning phases """
        return [0] + np.cumsum([p.number_of_epochs for p in phases]).tolist()[:-1]

    def _select_phase_left_bound(self, epoch_number):
        """
        Return number of current phase.
        Return index of first phase not done after all up to epoch_number were done.
        """
        idx = bisect.bisect_left(self.ladder, epoch_number)

        if idx >= len(self.ladder):
            return len(self.ladder) - 1
        elif self.ladder[idx] > epoch_number:
            return idx - 1
        else:
            return idx

    def _select_phase_right_bound(self, epoch_number):
        """
        Return number of current phase.
        Return index of first phase not done after all up to epoch_number were done.
        """
        return bisect.bisect_right(self.ladder, epoch_number) - 1

    def run(self):
        """ Run the command with supplied configuration """
        device = torch.device(self.model_config.device)
        learner = Learner(device, self.model_factory.instantiate())

        # All callbacks useful for learning
        callbacks = self.gather_callbacks()

        # Metrics to track through this training
        metrics = learner.metrics()

        # Check if training was already started and potentially continue where we left off
        training_info, hidden_state = self.resume_training(learner, callbacks, metrics)

        # Prepare current training phase
        current_phase_idx = self._select_phase_left_bound(training_info.start_epoch_idx)
        current_phase = self.phases[current_phase_idx]
        local_idx = training_info.start_epoch_idx - self.ladder[current_phase_idx]

        current_phase.set_up_phase(training_info, learner.model, self.source)
        print(current_phase.banner())

        if training_info.start_epoch_idx > 0:
            current_phase.restore(training_info, local_idx, learner.model, hidden_state)

        training_info.on_train_begin()

        for global_epoch_idx in range(training_info.start_epoch_idx + 1, self.full_number_of_epochs + 1):
            iteration_phase_idx = self._select_phase_right_bound(global_epoch_idx-1)
            local_idx = global_epoch_idx - self.ladder[iteration_phase_idx]

            # Phase preparations
            while current_phase_idx != iteration_phase_idx:
                current_phase.tear_down_phase(training_info, learner.model)

                current_phase_idx += 1
                current_phase = self.phases[current_phase_idx]

                current_phase.set_up_phase(training_info, learner.model, self.source)
                print(current_phase.banner())

            # Create epoch info
            epoch_info = current_phase.epoch_info(training_info, global_epoch_idx, local_idx)

            # Execute learning
            current_phase.execute_epoch(epoch_info, learner)

            # Epoch checkpoint
            self.storage.checkpoint(epoch_info, learner.model)

        # Tear down the last phase
        if current_phase is not None:
            current_phase.tear_down_phase(training_info, learner.model)

        training_info.on_train_end()

        return training_info

    def gather_callbacks(self) -> list:
        """ Gather all the callbacks to be used in this training run """
        callbacks = []

        callbacks.extend(self.callbacks)
        callbacks.extend(self.storage.streaming_callbacks())

        return callbacks

    def resume_training(self, learner, callbacks, metrics) -> (TrainingInfo, dict):
        """ Possibly resume training from a saved state from the storage """
        if self.model_config.continue_training:
            start_epoch = self.storage.last_epoch_idx()
        else:
            start_epoch = 0

        training_info = TrainingInfo(
            start_epoch_idx=start_epoch,
            run_name=self.model_config.run_name,
            metrics=metrics,
            callbacks=callbacks
        )

        if start_epoch == 0:
            self.storage.reset(self.model_config.render_configuration())
            training_info.initialize()
            learner.initialize_training(training_info)
            hidden_state = None
        else:
            hidden_state = self.storage.resume(training_info, learner.model)

        return training_info, hidden_state


def create(model_config, model, source, storage, phases, callbacks=None, restart=True):
    """ Vel creation function """
    return PhaseTrainCommand(
        model_config=model_config,
        model_factory=model,
        source=source,
        storage=storage,
        phases=phases,
        callbacks=callbacks,
        restart=restart
    )





class Augmentation:
    """ Base class for all data augmentations """
    def __init__(self, mode='x', tags=None):
        self.mode = mode
        self.tags = tags or ['train', 'val', 'test']

    def __call__(self, *args):
        """ Do the transformation """
        print(self)
        raise NotImplementedError

    def denormalize(self, *args):
        """ Operation reverse to normalization """
        if len(args) == 1:
            return args[0]
        else:
            return args



from .model import Model


class ModelFactory:
    """ Factory class for models """

    def instantiate(self, **extra_args) -> Model:
        raise NotImplementedError

    @staticmethod
    def generic(closure):
        return GenericModelFactory(closure)


class GenericModelFactory(ModelFactory):
    """ Create model from a lambda function """
    def __init__(self, closure):
        self.closure = closure

    def instantiate(self, **extra_args):
        return self.closure(**extra_args)





def divide_ceiling(numerator, denominator):
    """ Determine the smallest number k such, that denominator * k >= numerator """
    split_val = numerator // denominator
    rest = numerator % denominator

    if rest > 0:
        return split_val + 1
    else:
        return split_val



import re
import os.path as osp
import os
SCRIPT_DIR = os.path.dirname(os.path.abspath(__file__))

_atari7 = ['BeamRider', 'Breakout', 'Enduro', 'Pong', 'Qbert', 'Seaquest', 'SpaceInvaders']
_atariexpl7 = ['Freeway', 'Gravitar', 'MontezumaRevenge', 'Pitfall', 'PrivateEye', 'Solaris', 'Venture']

_BENCHMARKS = []

remove_version_re = re.compile(r'-v\d+$')


def register_benchmark(benchmark):
    for b in _BENCHMARKS:
        if b['name'] == benchmark['name']:
            raise ValueError('Benchmark with name %s already registered!' % b['name'])

    # automatically add a description if it is not present
    if 'tasks' in benchmark:
        for t in benchmark['tasks']:
            if 'desc' not in t:
                t['desc'] = remove_version_re.sub('', t['env_id'])
    _BENCHMARKS.append(benchmark)


def list_benchmarks():
    return [b['name'] for b in _BENCHMARKS]


def get_benchmark(benchmark_name):
    for b in _BENCHMARKS:
        if b['name'] == benchmark_name:
            return b
    raise ValueError('%s not found! Known benchmarks: %s' % (benchmark_name, list_benchmarks()))


def get_task(benchmark, env_id):
    """Get a task by env_id. Return None if the benchmark doesn't have the env"""
    return next(filter(lambda task: task['env_id'] == env_id, benchmark['tasks']), None)


def find_task_for_env_id_in_any_benchmark(env_id):
    for bm in _BENCHMARKS:
        for task in bm["tasks"]:
            if task["env_id"] == env_id:
                return bm, task
    return None, None


_ATARI_SUFFIX = 'NoFrameskip-v4'

register_benchmark({
    'name': 'Atari50M',
    'description': '7 Atari games from Mnih et al. (2013), with pixel observations, 50M timesteps',
    'tasks': [{'desc': _game, 'env_id': _game + _ATARI_SUFFIX, 'trials': 2, 'num_timesteps': int(50e6)} for _game in _atari7]
})

register_benchmark({
    'name': 'Atari10M',
    'description': '7 Atari games from Mnih et al. (2013), with pixel observations, 10M timesteps',
    'tasks': [{'desc': _game, 'env_id': _game + _ATARI_SUFFIX, 'trials': 2, 'num_timesteps': int(10e6)} for _game in _atari7]
})

register_benchmark({
    'name': 'Atari1Hr',
    'description': '7 Atari games from Mnih et al. (2013), with pixel observations, 1 hour of walltime',
    'tasks': [{'desc': _game, 'env_id': _game + _ATARI_SUFFIX, 'trials': 2, 'num_seconds': 60 * 60} for _game in _atari7]
})

register_benchmark({
    'name': 'AtariExploration10M',
    'description': '7 Atari games emphasizing exploration, with pixel observations, 10M timesteps',
    'tasks': [{'desc': _game, 'env_id': _game + _ATARI_SUFFIX, 'trials': 2, 'num_timesteps': int(10e6)} for _game in _atariexpl7]
})


# MuJoCo

_mujocosmall = [
    'InvertedDoublePendulum-v2', 'InvertedPendulum-v2',
    'HalfCheetah-v2', 'Hopper-v2', 'Walker2d-v2',
    'Reacher-v2', 'Swimmer-v2']
register_benchmark({
    'name': 'Mujoco1M',
    'description': 'Some small 2D MuJoCo tasks, run for 1M timesteps',
    'tasks': [{'env_id': _envid, 'trials': 3, 'num_timesteps': int(1e6)} for _envid in _mujocosmall]
})
register_benchmark({
    'name': 'MujocoWalkers',
    'description': 'MuJoCo forward walkers, run for 8M, humanoid 100M',
    'tasks': [
        {'env_id': "Hopper-v1", 'trials': 4, 'num_timesteps': 8 * 1000000},
        {'env_id': "Walker2d-v1", 'trials': 4, 'num_timesteps': 8 * 1000000},
        {'env_id': "Humanoid-v1", 'trials': 4, 'num_timesteps': 100 * 1000000},
    ]
})

# Roboschool

register_benchmark({
    'name': 'Roboschool8M',
    'description': 'Small 2D tasks, up to 30 minutes to complete on 8 cores',
    'tasks': [
        {'env_id': "RoboschoolReacher-v1", 'trials': 4, 'num_timesteps': 2 * 1000000},
        {'env_id': "RoboschoolAnt-v1", 'trials': 4, 'num_timesteps': 8 * 1000000},
        {'env_id': "RoboschoolHalfCheetah-v1", 'trials': 4, 'num_timesteps': 8 * 1000000},
        {'env_id': "RoboschoolHopper-v1", 'trials': 4, 'num_timesteps': 8 * 1000000},
        {'env_id': "RoboschoolWalker2d-v1", 'trials': 4, 'num_timesteps': 8 * 1000000},
    ]
})
register_benchmark({
    'name': 'RoboschoolHarder',
    'description': 'Test your might!!! Up to 12 hours on 32 cores',
    'tasks': [
        {'env_id': "RoboschoolHumanoid-v1", 'trials': 4, 'num_timesteps': 100 * 1000000},
        {'env_id': "RoboschoolHumanoidFlagrun-v1", 'trials': 4, 'num_timesteps': 200 * 1000000},
        {'env_id': "RoboschoolHumanoidFlagrunHarder-v1", 'trials': 4, 'num_timesteps': 400 * 1000000},
    ]
})

# Other

_atari50 = [  # actually 47
    'Alien', 'Amidar', 'Assault', 'Asterix', 'Asteroids',
    'Atlantis', 'BankHeist', 'BattleZone', 'BeamRider', 'Bowling',
    'Breakout', 'Centipede', 'ChopperCommand', 'CrazyClimber',
    'DemonAttack', 'DoubleDunk', 'Enduro', 'FishingDerby', 'Freeway',
    'Frostbite', 'Gopher', 'Gravitar', 'IceHockey', 'Jamesbond',
    'Kangaroo', 'Krull', 'KungFuMaster', 'MontezumaRevenge', 'MsPacman',
    'NameThisGame', 'Pitfall', 'Pong', 'PrivateEye', 'Qbert',
    'RoadRunner', 'Robotank', 'Seaquest', 'SpaceInvaders', 'StarGunner',
    'Tennis', 'TimePilot', 'Tutankham', 'UpNDown', 'Venture',
    'VideoPinball', 'WizardOfWor', 'Zaxxon',
]

register_benchmark({
    'name': 'Atari50_10M',
    'description': '47 Atari games from Mnih et al. (2013), with pixel observations, 10M timesteps',
    'tasks': [{'desc': _game, 'env_id': _game + _ATARI_SUFFIX, 'trials': 2, 'num_timesteps': int(10e6)} for _game in _atari50]
})

# HER DDPG

register_benchmark({
    'name': 'HerDdpg',
    'description': 'Smoke-test only benchmark of HER',
    'tasks': [{'trials': 1, 'env_id': 'FetchReach-v1'}]
})



import torch.optim.lr_scheduler as scheduler
import numpy as np

import vel.api.base as base


class LadderScheduler(base.Callback):
    """ Scheduler defined by a set of learning rates after reaching given number of iterations """
    def __init__(self, optimizer, ladder, last_epoch):
        self.schedule_limits = np.cumsum([x[0] for x in ladder])
        self.schedule_numbers = np.array([float(x[1]) for x in ladder])
        self.scheduler = scheduler.LambdaLR(optimizer, self.lambda_fn, last_epoch=last_epoch)

    def lambda_fn(self, epoch_idx):
        idx = np.minimum(np.searchsorted(self.schedule_limits, epoch_idx), len(self.schedule_limits) - 1)
        return self.schedule_numbers[idx]

    def on_epoch_begin(self, epoch_info):
        self.scheduler.step(epoch=epoch_info.global_epoch_idx)


class LadderSchedulerFactory(base.SchedulerFactory):
    """ Factory class for ladder scheduler """
    def __init__(self, ladder):
        self.ladder = ladder

    def instantiate(self, optimizer, last_epoch=-1) -> LadderScheduler:
        return LadderScheduler(optimizer, self.ladder, last_epoch)


def create(ladder):
    """ Create a hand-scheduled ladder scheduler """
    return LadderSchedulerFactory(ladder)



import gym
import numpy as np

from vel.exceptions import VelException


def take_along_axis(large_array, indexes):
    """ Take along axis """
    # Reshape indexes into the right shape
    if len(large_array.shape) > len(indexes.shape):
        indexes = indexes.reshape(indexes.shape + tuple([1] * (len(large_array.shape) - len(indexes.shape))))

    return np.take_along_axis(large_array, indexes, axis=0)


class DequeMultiEnvBufferBackend:
    """
    Simple backend behind DequeBuffer - version supporting multiple environments.

    Frame stack compensation - if environment has a framestack built in, we will store only the last action
    """

    def __init__(self, buffer_capacity: int, num_envs: int, observation_space: gym.Space, action_space: gym.Space,
                 extra_data=None, frame_stack_compensation: bool=False):
        # Maximum number of items in the buffer
        self.buffer_capacity = buffer_capacity

        self.frame_stack_compensation = frame_stack_compensation

        # Number of parallel envs to record
        self.num_envs = num_envs

        # How many elements have been inserted in the buffer
        self.current_size = 0

        # Index of last inserted element
        self.current_idx = -1

        # Data buffers
        if self.frame_stack_compensation:
            self.state_buffer = np.zeros(
                [self.buffer_capacity, self.num_envs] + list(observation_space.shape)[:-1] + [1],
                dtype=observation_space.dtype
            )
        else:
            self.state_buffer = np.zeros(
                [self.buffer_capacity, self.num_envs] + list(observation_space.shape),
                dtype=observation_space.dtype
            )

        self.action_buffer = np.zeros(
            [self.buffer_capacity, self.num_envs] + list(action_space.shape), dtype=action_space.dtype
        )
        self.reward_buffer = np.zeros([self.buffer_capacity, self.num_envs], dtype=np.float32)
        self.dones_buffer = np.zeros([self.buffer_capacity, self.num_envs], dtype=bool)

        # One list per environment
        self.extra_data = {} if extra_data is None else extra_data

        # Just a sentinel to simplify further calculations
        self.dones_buffer[self.current_idx] = True

    def store_transition(self, frame, action, reward, done, extra_info=None):
        """ Store given transition in the backend """
        self.current_idx = (self.current_idx + 1) % self.buffer_capacity

        if self.frame_stack_compensation:
            # Compensate for frame stack built into the environment
            frame = np.expand_dims(np.take(frame, indices=-1, axis=-1), axis=-1)

        self.state_buffer[self.current_idx] = frame

        self.action_buffer[self.current_idx] = action
        self.reward_buffer[self.current_idx] = reward
        self.dones_buffer[self.current_idx] = done

        for name in self.extra_data:
            self.extra_data[name][self.current_idx] = extra_info[name]

        if self.current_size < self.buffer_capacity:
            self.current_size += 1

        return self.current_idx

    def get_frame_with_future(self, frame_idx, env_idx, history_length=1):
        """ Return frame from the buffer together with the next frame """
        if frame_idx == self.current_idx:
            raise VelException("Cannot provide enough future for the frame")

        past_frame = self.get_frame(frame_idx, env_idx, history_length)

        if history_length > 1:
            assert self.state_buffer.shape[-1] == 1, \
                "State buffer must have last dimension of 1 if we want frame history"

        if not self.dones_buffer[frame_idx, env_idx]:
            next_idx = (frame_idx + 1) % self.buffer_capacity
            next_frame = self.state_buffer[next_idx, env_idx]
        else:
            next_idx = (frame_idx + 1) % self.buffer_capacity
            next_frame = np.zeros_like(self.state_buffer[next_idx, env_idx])

        if history_length > 1:
            future_frame = np.concatenate([
                past_frame.take(indices=np.arange(1, past_frame.shape[-1]), axis=-1), next_frame
            ], axis=-1)
        else:
            future_frame = next_frame

        return past_frame, future_frame

    def get_frame(self, frame_idx, env_idx, history_length=1):
        """ Return frame from the buffer """
        if frame_idx >= self.current_size:
            raise VelException("Requested frame beyond the size of the buffer")

        if history_length > 1:
            assert self.state_buffer.shape[-1] == 1, "State buffer must have last dimension of 1 if we want frame history"

        accumulator = []

        last_frame = self.state_buffer[frame_idx, env_idx]

        accumulator.append(last_frame)

        for i in range(history_length - 1):
            prev_idx = (frame_idx - 1) % self.buffer_capacity

            if prev_idx == self.current_idx:
                raise VelException("Cannot provide enough history for the frame")
            elif self.dones_buffer[prev_idx, env_idx]:
                # If previous frame was done - just append zeroes
                accumulator.append(np.zeros_like(last_frame))
            else:
                frame_idx = prev_idx
                accumulator.append(self.state_buffer[frame_idx, env_idx])

        # We're pushing the elements in reverse order
        return np.concatenate(accumulator[::-1], axis=-1)

    def get_transition(self, frame_idx, env_idx, history_length=1):
        """ Single transition with given index """
        past_frame, future_frame = self.get_frame_with_future(frame_idx, env_idx, history_length)

        data_dict = {
            'state': past_frame,
            'state+1': future_frame,
            'action': self.action_buffer[frame_idx, env_idx],
            'reward': self.reward_buffer[frame_idx, env_idx],
            'done': self.dones_buffer[frame_idx, env_idx],
        }

        for name in self.extra_data:
            data_dict[name] = self.extra_data[name][frame_idx, env_idx]

        return data_dict

    def get_batch(self, indexes, history_length):
        """ Return batch with given indexes """
        assert indexes.shape[1] == self.state_buffer.shape[1], \
            "Must have the same number of indexes as there are environments"

        frame_batch_shape = (
                [indexes.shape[0], indexes.shape[1]]
                + list(self.state_buffer.shape[2:-1])
                + [self.state_buffer.shape[-1] * history_length]
        )

        past_frame_buffer = np.zeros(frame_batch_shape, dtype=self.state_buffer.dtype)
        future_frame_buffer = np.zeros(frame_batch_shape, dtype=self.state_buffer.dtype)

        for buffer_idx, frame_row in enumerate(indexes):
            for env_idx, frame_idx in enumerate(frame_row):
                past_frame_buffer[buffer_idx, env_idx], future_frame_buffer[buffer_idx, env_idx] = (
                    self.get_frame_with_future(frame_idx, env_idx, history_length)
                )

        actions = take_along_axis(self.action_buffer, indexes)
        rewards = take_along_axis(self.reward_buffer, indexes)
        dones = take_along_axis(self.dones_buffer, indexes)

        data_dict = {
            'states': past_frame_buffer,
            'actions': actions,
            'rewards': rewards,
            'states+1': future_frame_buffer,
            'dones': dones,
        }

        for name in self.extra_data:
            data_dict[name] = take_along_axis(self.extra_data[name], indexes)

        return data_dict

    def sample_batch_uniform(self, batch_size, history_length):
        """ Return indexes of next sample"""
        results = []

        for i in range(self.num_envs):
            results.append(self.sample_uniform_single_env(batch_size, history_length))

        return np.stack(results, axis=-1)

    def sample_batch_rollout(self, rollout_length, history_length):
        """ Return indexes of next random rollout """
        results = []

        for i in range(self.num_envs):
            results.append(self.sample_rollout_single_env(rollout_length, history_length))

        return np.stack(results, axis=-1)

    def get_rollout(self, indexes, rollout_length, history_length):
        """ Return batch consisting of *consecutive* transitions """
        assert indexes.shape[0] > 1, "There must be multiple indexes supplied"
        assert rollout_length > 1, "Rollout length must be greater than 1"

        batch_indexes = indexes.reshape(1, indexes.shape[0]) - np.arange(rollout_length - 1, -1, -1).reshape(rollout_length, 1)

        return self.get_batch(batch_indexes, history_length)

    def sample_rollout_single_env(self, rollout_length, history_length):
        """ Return indexes of next sample"""
        # Sample from up to total size
        if self.current_size < self.buffer_capacity:
            if rollout_length + 1 > self.current_size:
                raise VelException("Not enough elements in the buffer to sample the rollout")

            # -1 because we cannot take the last one
            return np.random.choice(self.current_size - rollout_length) + rollout_length - 1
        else:
            if rollout_length + history_length > self.current_size:
                raise VelException("Not enough elements in the buffer to sample the rollout")

            candidate = np.random.choice(self.buffer_capacity)

            # These are the elements we cannot draw, as then we don't have enough history
            forbidden_ones = (
                    np.arange(self.current_idx, self.current_idx + history_length + rollout_length - 1) % self.buffer_capacity
            )

            # Exclude these frames for learning as they may have some part of history overwritten
            while candidate in forbidden_ones:
                candidate = np.random.choice(self.buffer_capacity)

            return candidate

    def sample_uniform_single_env(self, batch_size, history_length):
        """ Return indexes of next sample"""
        # Sample from up to total size
        if self.current_size < self.buffer_capacity:
            # -1 because we cannot take the last one
            return np.random.choice(self.current_size - 1, batch_size, replace=False)
        else:
            candidate = np.random.choice(self.buffer_capacity, batch_size, replace=False)

            forbidden_ones = (
                    np.arange(self.current_idx, self.current_idx + history_length) % self.buffer_capacity
            )

            # Exclude these frames for learning as they may have some part of history overwritten
            while any(x in candidate for x in forbidden_ones):
                candidate = np.random.choice(self.buffer_capacity, batch_size, replace=False)

            return candidate



import certifi
import numpy as np
import os
import pathlib
import urllib3

import torch

from vel.api.base import Source


class TextIterator:
    """ Iterator over a text dataset """
    def __init__(self, padded_sequence, sequence_length, batch_size, alphabet_size, num_batches):
        self.sequence_length = sequence_length
        self.batch_size = batch_size
        self.num_batches = num_batches
        self.alphabet_size = alphabet_size

        self.padded_sequence = padded_sequence[:-1].reshape(self.num_batches * self.batch_size, self.sequence_length)
        self.padded_sequence_next = padded_sequence[1:].reshape(self.num_batches * self.batch_size, self.sequence_length)

        self.sequence_indices = np.arange(self.num_batches * self.batch_size)

        np.random.shuffle(self.sequence_indices)

        self.sequence_indices = self.sequence_indices.reshape(self.num_batches, self.batch_size)

        self.batch_idx = 0

    def __iter__(self):
        return self

    def __next__(self):
        if self.batch_idx == self.num_batches:
            raise StopIteration
        else:
            input_data = torch.from_numpy(self.padded_sequence[self.sequence_indices[self.batch_idx]])
            target_data = torch.from_numpy(self.padded_sequence_next[self.sequence_indices[self.batch_idx]])

            self.batch_idx += 1

            return input_data.to(torch.long), target_data.to(torch.long)


class TextLoader:
    """ Loader of sequential text data """
    def __init__(self, sequence, sequence_length, batch_size, alphabet_size):
        self.sequence = sequence
        self.sequence_length = sequence_length
        self.batch_size = batch_size
        self.alphabet_size = alphabet_size

        # 1 is for the last element as the target needs to be shifted by 1
        residual_length = (len(self.sequence) - self.sequence_length - 1)
        full_size = self.sequence_length * self.batch_size

        rest = residual_length % full_size
        self.num_batches = residual_length // full_size

        if rest > 0:
            self.sequence = np.pad(self.sequence, (0, full_size - rest), mode='constant')
            self.num_batches += 1

    def __iter__(self):
        initial_offset = np.random.randint(self.sequence_length)
        relevant_subsequence = self.sequence[
           # 1 is for the last element as the target needs to be shifted by 1
           initial_offset:self.num_batches * self.sequence_length * self.batch_size + initial_offset + 1
        ]

        return TextIterator(
            relevant_subsequence, self.sequence_length, self.batch_size,
            alphabet_size=self.alphabet_size,
            num_batches=self.num_batches
        )

    def __len__(self):
        """ Number of batches in this loader """
        return self.num_batches


class TextUrlSource(Source):
    """ Download text from source and model it character by character """
    def __init__(self, url, absolute_data_path, sequence_length, batch_size, train_val_split=0.8):
        super().__init__()

        self.url = url
        self.data_path = absolute_data_path
        self.sequence_length = sequence_length
        self.batch_size = batch_size
        self.train_val_split = train_val_split

        self.text_path = os.path.join(self.data_path, 'text.txt')
        self.processed_path = os.path.join(self.data_path, 'processed.data')

        self.data_dict = self.download()

        content_encoded = self.data_dict['content_encoded']
        alphabet_size = len(self.data_dict['alphabet'])

        split_idx = int(len(content_encoded) * train_val_split)

        self._train_loader = TextLoader(
            sequence=content_encoded[:split_idx],
            sequence_length=sequence_length,
            batch_size=batch_size,
            alphabet_size=alphabet_size,
        )

        self._val_loader = TextLoader(
            sequence=content_encoded[split_idx:],
            sequence_length=sequence_length,
            batch_size=batch_size,
            alphabet_size=alphabet_size,
        )

    def encode_character(self, char):
        return self.data_dict['character_to_index'][char]

    def decode_character(self, index):
        return self.data_dict['index_to_character'][index]

    def train_loader(self):
        """ PyTorch loader of training data """
        return self._train_loader

    def val_loader(self):
        """ PyTorch loader of validation data """
        return self._val_loader

    def train_dataset(self):
        """ Return the training dataset """
        return None

    def val_dataset(self):
        """ Return the validation dataset """
        return None

    def train_iterations_per_epoch(self):
        """ Return number of iterations per epoch """
        return len(self._train_loader)

    def val_iterations_per_epoch(self):
        """ Return number of iterations per epoch - validation """
        return len(self._val_loader)

    def download(self):
        """ Make sure data file is downloaded and stored properly """
        if not os.path.exists(self.data_path):
            # Create if it doesn't exist
            pathlib.Path(self.data_path).mkdir(parents=True, exist_ok=True)

        if not os.path.exists(self.text_path):
            http = urllib3.PoolManager(cert_reqs='CERT_REQUIRED', ca_certs=certifi.where())

            with open(self.text_path, 'wt') as fp:
                request = http.request('GET', self.url)
                content = request.data.decode('utf8')
                fp.write(content)

        if not os.path.exists(self.processed_path):
            with open(self.text_path, 'rt') as fp:
                content = fp.read()

            alphabet = sorted(set(content))

            index_to_character = {idx: c for idx, c in enumerate(alphabet, 1)}
            character_to_index = {c: idx for idx, c in enumerate(alphabet, 1)}

            content_encoded = np.array([character_to_index[c] for c in content], dtype=np.uint8)

            data_dict = {
                'alphabet': alphabet,
                'index_to_character': index_to_character,
                'character_to_index': character_to_index,
                'content_encoded': content_encoded
            }

            with open(self.processed_path, 'wb') as fp:
                torch.save(data_dict, fp)
        else:
            with open(self.processed_path, 'rb') as fp:
                data_dict = torch.load(fp)

        return data_dict


def create(model_config, url, local_dir, sequence_length=64, batch_size=64, train_val_split=0.8):
    """ Vel creation function """
    if not os.path.isabs(local_dir):
        local_dir = model_config.project_data_dir(local_dir)

    return TextUrlSource(
        url,
        absolute_data_path=local_dir,
        sequence_length=sequence_length,
        batch_size=batch_size,
        train_val_split=train_val_split,
    )



import matplotlib.pyplot as plt
import numpy as np

from vel.api.base import Source


class AugmentationVisualizationCommand:
    """ Visualize augmentations """
    def __init__(self, source: Source, samples, cases):
        self.source = source
        self.samples = samples
        self.cases = cases

    def run(self):
        """ Run the visualization """
        dataset = self.source.train_dataset()
        num_samples = len(dataset)

        fig, ax = plt.subplots(self.cases, self.samples+1)

        selected_sample = np.sort(np.random.choice(num_samples, self.cases, replace=False))

        for i in range(self.cases):
            raw_image, _ = dataset.get_raw(selected_sample[i])

            ax[i, 0].imshow(raw_image)
            ax[i, 0].set_title("Original image")

            for j in range(self.samples):
                augmented_image, _ = dataset[selected_sample[i]]
                augmented_image = dataset.denormalize(augmented_image)
                ax[i, j+1].imshow(augmented_image)

        plt.show()


def create(source, samples, cases):
    """ Visualize augmentations in the source """
    return AugmentationVisualizationCommand(source, samples, cases)



import importlib
import inspect

from vel.internals.parser import Variable


class Provider:
    """ Dependency injection resolver for the configuration file """
    def __init__(self, environment, instances=None, parameters=None):
        self.environment = environment

        self.parameters = parameters if parameters is not None else {}

        self.instances = {
            **(instances if instances is not None else {}),
            'vel_provider': self
        }

    def inject(self, name, value):
        """ Inject an object into the provider """
        self.instances[name] = value

    def resolve_and_call(self, func, extra_env=None):
        """ Resolve function arguments and call them, possibily filling from the environment """
        parameter_list = [
            (k, v.default == inspect.Parameter.empty) for k, v in inspect.signature(func).parameters.items()
        ]
        extra_env = extra_env if extra_env is not None else {}
        kwargs = {}

        for parameter_name, is_required in parameter_list:
            if parameter_name in extra_env:
                kwargs[parameter_name] = self.instantiate_from_data(extra_env[parameter_name])
                continue

            if parameter_name in self.instances:
                kwargs[parameter_name] = self.instances[parameter_name]
                continue

            if parameter_name in self.environment:
                kwargs[parameter_name] = self.instantiate_by_name(parameter_name)
                continue

            if is_required:
                funcname = f"{inspect.getmodule(func).__name__}.{func.__name__}"
                raise RuntimeError("Required argument '{}' cannot be resolved for function '{}'".format(
                    parameter_name, funcname
                ))

        return func(**kwargs)

    def instantiate_from_data(self, object_data):
        """ Instantiate object from the supplied data, additional args may come from the environment """
        if isinstance(object_data, dict) and 'name' in object_data:
            name = object_data['name']
            module = importlib.import_module(name)
            return self.resolve_and_call(module.create, extra_env=object_data)
        elif isinstance(object_data, dict):
            return {k: self.instantiate_from_data(v) for k, v in object_data.items()}
        elif isinstance(object_data, list):
            return [self.instantiate_from_data(x) for x in object_data]
        elif isinstance(object_data, Variable):
            return object_data.resolve(self.parameters)
        else:
            return object_data

    def render_configuration(self, configuration=None):
        """ Render variables in configuration object but don't instantiate anything """
        if configuration is None:
            configuration = self.environment

        if isinstance(configuration, dict):
            return {k: self.render_configuration(v) for k, v in configuration.items()}
        elif isinstance(configuration, list):
            return [self.render_configuration(x) for x in configuration]
        elif isinstance(configuration, Variable):
            return configuration.resolve(self.parameters)
        else:
            return configuration

    def has_name(self, object_name):
        """ Check if given name is available in the provider """
        return object_name in self.instances or object_name in self.environment

    def instantiate_by_name(self, object_name):
        """ Instantiate object from the environment, possibly giving some extra arguments """
        if object_name not in self.instances:
            instance = self.instantiate_from_data(self.environment[object_name])

            self.instances[object_name] = instance
            return instance
        else:
            return self.instances[object_name]



import torch

from vel.api.base import Model
from vel.api import TrainingInfo, EpochInfo, BatchInfo


class ReinforcerBase:
    """
    Manages training process of a single model.
    Learner version for reinforcement-learning problems.
    """

    def initialize_training(self, training_info: TrainingInfo):
        """ Run the initialization procedure """
        pass

    def train_epoch(self, epoch_info: EpochInfo):
        """ Train model on an epoch of a fixed number of batch updates """
        raise NotImplementedError

    def train_batch(self, batch_info: BatchInfo):
        """ Single, most atomic 'step' of learning this reinforcer can perform """
        raise NotImplementedError

    def metrics(self) -> list:
        """ List of metrics to track for this learning process """
        raise NotImplementedError

    @property
    def model(self) -> Model:
        """ Model trained by this reinforcer """
        raise NotImplementedError


class ReinforcerFactory:
    """ A reinforcer factory """
    def instantiate(self, device: torch.device) -> ReinforcerBase:
        """ Create new reinforcer instance """
        raise NotImplementedError



import torch.nn as nn
import torch.nn.init as init


class DeterministicCriticHead(nn.Module):
    """
    Network head for action-dependent critic.
    Returns deterministic action-value for given combination of action and state.
    """

    def __init__(self, input_dim):
        super().__init__()

        self.input_dim = input_dim
        self.linear = nn.Linear(input_dim, 1)

    def forward(self, input_data):
        return self.linear(input_data)[:, 0]

    def reset_weights(self):
        """ Initialize weights to sane defaults """
        init.uniform_(self.linear.weight, -3e-3, 3e-3)
        init.zeros_(self.linear.bias)



import time

from vel.api import BatchInfo, TrainingInfo
from vel.api.base import Callback


class TimeTracker(Callback):
    """ Track training time - in seconds """
    def __init__(self):
        self.start_time = None

    def on_initialization(self, training_info: TrainingInfo):
        training_info['time'] = 0.0

    def on_train_begin(self, training_info: TrainingInfo):
        self.start_time = time.time()

    def on_batch_end(self, batch_info: BatchInfo):
        current_time = time.time()
        batch_time = current_time - self.start_time
        self.start_time = current_time

        batch_info['time'] = batch_time
        batch_info.training_info['time'] += batch_info['time']

    def write_state_dict(self, training_info: TrainingInfo, hidden_state_dict: dict):
        hidden_state_dict['time_tracker/time'] = training_info['time']

    def load_state_dict(self, training_info: TrainingInfo, hidden_state_dict: dict):
        training_info['time'] = hidden_state_dict['time_tracker/time']



import datetime as dtm
import os.path
import torch

from vel.exceptions import VelInitializationException
from vel.internals.parser import Parser
from vel.internals.provider import Provider


class ModelConfig:
    """
    Read from YAML configuration of a model, specifying all details of the run.
    Is a frontend for the provider, resolving all dependency-injection requests.
    """

    PROJECT_FILE_NAME = '.velproject.yaml'

    @staticmethod
    def find_project_directory(start_path) -> str:
        """ Locate top-level project directory  """
        start_path = os.path.realpath(start_path)
        possible_name = os.path.join(start_path, ModelConfig.PROJECT_FILE_NAME)

        if os.path.exists(possible_name):
            return start_path
        else:
            up_path = os.path.realpath(os.path.join(start_path, '..'))
            if os.path.realpath(start_path) == up_path:
                raise RuntimeError(f"Couldn't find project file starting from {start_path}")
            else:
                return ModelConfig.find_project_directory(up_path)

    @classmethod
    def from_file(cls, filename: str, run_number: int, continue_training=False, seed: int=None, device: str='cuda', params=None):
        """ Create model config from file """
        with open(filename, 'r') as fp:
            model_config_contents = Parser.parse(fp)

        project_config_path = ModelConfig.find_project_directory(os.path.dirname(os.path.abspath(filename)))

        with open(os.path.join(project_config_path, cls.PROJECT_FILE_NAME), 'r') as fp:
            project_config_contents = Parser.parse(fp)

        aggregate_dictionary = {
            **project_config_contents,
            **model_config_contents
        }

        # Options that should exist for every config
        try:
            model_name = model_config_contents['name']
        except KeyError:
            raise VelInitializationException("Model configuration must have a 'name' key")

        return ModelConfig(
            model_name=model_name,
            filename=filename,
            configuration=aggregate_dictionary,
            run_number=run_number,
            project_dir=project_config_path,
            continue_training=continue_training,
            seed=seed,
            device=device,
            parameters=params
        )

    @classmethod
    def from_memory(cls, model_name: str, model_data: dict, run_number: int, project_dir: str,
                    continue_training=False, seed: int=None, device: str='cuda', params=None):
        """ Create model config from supplied data """
        return ModelConfig(
            model_name=model_name,
            filename="[memory]",
            configuration=model_data,
            run_number=run_number,
            project_dir=project_dir,
            continue_training=continue_training,
            seed=seed,
            device=device,
            parameters=params
        )

    def __init__(self, model_name: str, filename: str, configuration: dict, run_number: int, project_dir: str,
                 continue_training=False, seed: int=None, device: str= 'cuda', parameters=None):
        self._model_name = model_name
        self.filename = filename
        self.device = device
        self.continue_training = continue_training
        self.run_number = run_number
        self.seed = seed if seed is not None else (dtm.date.today().year + self.run_number)

        self.contents = configuration
        self.project_dir = project_dir

        self.command_descriptors = self.contents.get('commands', [])

        # This one is special and needs to get removed
        if 'commands' in self.contents:
            del self.contents['commands']

        self.provider = Provider(self._prepare_environment(), {'model_config': self}, parameters=parameters)

    def _prepare_environment(self) -> dict:
        """ Return full environment for dependency injection """
        return {**self.contents, 'run_number': self.run_number}

    def render_configuration(self) -> dict:
        """ Return a nice and picklable run configuration """
        return self.provider.render_configuration()

    ####################################################################################################################
    # COMMAND UTILITIES
    def get_command(self, command_name):
        """ Return object for given command """
        return self.provider.instantiate_from_data(self.command_descriptors[command_name])

    def run_command(self, command_name, varargs):
        """ Instantiate model class """
        command_descriptor = self.get_command(command_name)
        return command_descriptor.run(*varargs)

    ####################################################################################################################
    # MODEL DIRECTORIES
    def checkpoint_dir(self, *args) -> str:
        """ Return checkpoint directory for this model """
        return self.output_dir('checkpoints', self.run_name, *args)

    def data_dir(self, *args) -> str:
        """ Return data directory for given dataset """
        return self.project_data_dir(*args)

    def openai_dir(self) -> str:
        """ Return directory for openai output files for this model """
        return self.output_dir('openai', self.run_name)

    def project_data_dir(self, *args) -> str:
        """ Directory where to store data """
        return os.path.normpath(os.path.join(self.project_dir, 'data', *args))

    def output_dir(self, *args) -> str:
        """ Directory where to store output """
        return os.path.join(self.project_dir, 'output', *args)

    def project_top_dir(self, *args) -> str:
        """ Project top-level directory """
        return os.path.join(self.project_dir, *args)

    ####################################################################################################################
    # NAME UTILITIES
    @property
    def run_name(self) -> str:
        """ Return name of the run """
        return "{}/{}".format(self._model_name, self.run_number)

    @property
    def name(self) -> str:
        """ Return name of the model """
        return self._model_name

    ####################################################################################################################
    # PROVIDER API
    def provide(self, name):
        """ Return a dependency-injected instance """
        return self.provider.instantiate_by_name(name)

    ####################################################################################################################
    # BANNERS - Maybe shouldn't be here, but they are for now
    def banner(self, command_name) -> None:
        """ Print a banner for running the system """
        device = torch.device(self.device)
        print("=" * 80)
        print(f"Pytorch version: {torch.__version__} cuda version {torch.version.cuda} cudnn version {torch.backends.cudnn.version()}")
        print("Running model {}, run {} -- command {} -- device {}".format(self._model_name, self.run_number, command_name, self.device))
        if device.type == 'cuda':
            device_idx = 0 if device.index is None else device.index
            print(f"CUDA Device name {torch.cuda.get_device_name(device_idx)}")
        print(dtm.datetime.now().strftime("%Y/%m/%d - %H:%M:%S"))
        print("=" * 80)

    def quit_banner(self) -> None:
        """ Print a banner for running the system """
        print("=" * 80)
        print("Done.")
        print(dtm.datetime.now().strftime("%Y/%m/%d - %H:%M:%S"))
        print("=" * 80)

    ####################################################################################################################
    # Small UI utils
    def __repr__(self):
        return f"<ModelConfig at {self.filename}>"



import numpy as np
import torch

from vel.math.processes import OrnsteinUhlenbeckNoiseProcess
from vel.openai.baselines.common.running_mean_std import RunningMeanStd
from vel.rl.api import Rollout, Transitions
from vel.rl.api.base import ReplayEnvRollerBase, ReplayEnvRollerFactory
from vel.rl.buffers.deque_backend import DequeBufferBackend


class DequeReplayRollerOuNoise(ReplayEnvRollerBase):
    """
    Enrionment roller with experience replay buffer rolling out a **single** environment
    with Ornstein–Uhlenbeck noise process
    """

    def __init__(self, environment, device, batch_size, buffer_capacity, buffer_initial_size, noise_std_dev,
                 discount_factor, normalize_observations=False, normalize_returns=False):
        self.device = device
        self.batch_size = batch_size
        self.buffer_capacity = buffer_capacity
        self.buffer_initial_size = buffer_initial_size
        self.normalize_observations = normalize_observations
        self.normalize_returns = normalize_returns
        self.discount_factor = discount_factor

        self.device = device
        self._environment = environment

        self.backend = DequeBufferBackend(
            buffer_capacity=self.buffer_capacity,
            observation_space=environment.observation_space,
            action_space=environment.action_space
        )

        self.last_observation = self.environment.reset()

        len_action_space = self.environment.action_space.shape[-1]

        self.noise_process = OrnsteinUhlenbeckNoiseProcess(
            np.zeros(len_action_space), float(noise_std_dev) * np.ones(len_action_space)
        )

        self.ob_rms = RunningMeanStd(shape=self.environment.observation_space.shape) if normalize_observations else None
        self.ret_rms = RunningMeanStd(shape=()) if normalize_returns else None
        self.clip_obs = 5.0
        self.accumulated_return = 0.0

    @property
    def environment(self):
        """ Return environment of this env roller """
        return self._environment

    def is_ready_for_sampling(self) -> bool:
        """ If buffer is ready for drawing samples from it (usually checks if there is enough data) """
        return self.backend.current_size >= self.buffer_initial_size

    def _observation_to_tensor(self, observation_array):
        """ Convert observation numpy array to a tensor """
        return torch.from_numpy(self._filter_observation(observation_array)).to(self.device)

    @torch.no_grad()
    def rollout(self, batch_info, model) -> Rollout:
        """ Roll-out the environment and return it """
        observation_tensor = self._observation_to_tensor(self.last_observation[None])

        step = model.step(observation_tensor)
        action = step['actions'].detach().cpu().numpy()[0]
        noise = self.noise_process()

        action_perturbed = np.clip(
            action + noise, self.environment.action_space.low, self.environment.action_space.high
        )

        observation, reward, done, info = self.environment.step(action_perturbed)

        if self.ob_rms is not None:
            self.ob_rms.update(observation[None])

        if self.ret_rms is not None:
            self.accumulated_return = reward + self.discount_factor * self.accumulated_return

            self.ret_rms.update(np.array([self.accumulated_return]))

        self.backend.store_transition(self.last_observation, action_perturbed, reward, done)

        # Usual, reset on done
        if done:
            observation = self.environment.reset()
            self.noise_process.reset()
            self.accumulated_return = 0.0

        self.last_observation = observation

        return Transitions(
            size=1,
            environment_information=[info],
            transition_tensors={
                'actions': step['actions'],
                'values': step['values']
            },
        )

    def _filter_observation(self, obs):
        """ Potentially normalize observation """
        if self.ob_rms is not None:
            obs = np.clip((obs - self.ob_rms.mean) / np.sqrt(self.ob_rms.var + 1e-8), -self.clip_obs, self.clip_obs)

            return obs.astype(np.float32)
        else:
            return obs

    def sample(self, batch_info, model) -> Transitions:
        """ Sample experience from replay buffer and return a batch """
        indexes = self.backend.sample_batch_uniform(self.batch_size, history_length=1)
        batch = self.backend.get_batch(indexes, history_length=1)

        observations = self._observation_to_tensor(batch['states'])
        observations_plus1 = self._observation_to_tensor(batch['states+1'])

        rewards = batch['rewards'].astype(np.float32)

        if self.ret_rms is not None:
            rewards = np.clip(rewards / np.sqrt(self.ret_rms.var + 1e-8), -self.clip_obs, self.clip_obs)

        dones = torch.from_numpy(batch['dones'].astype(np.float32)).to(self.device)
        rewards = torch.from_numpy(rewards).to(self.device)
        actions = torch.from_numpy(batch['actions']).to(self.device)

        return Transitions(
            size=self.batch_size,
            environment_information=[],
            transition_tensors={
                'observations': observations,
                'observations_next': observations_plus1,
                'dones': dones,
                'rewards': rewards,
                'actions': actions
            }
        )


class DequeReplayRollerOuNoiseFactory(ReplayEnvRollerFactory):
    """ Factory class for DequeReplayQRoller """
    def __init__(self, buffer_capacity: int, buffer_initial_size: int, noise_std_dev: float,
                 normalize_observations: bool=False, normalize_returns: bool=False):
        self.buffer_capacity = buffer_capacity
        self.buffer_initial_size = buffer_initial_size
        self.noise_std_dev = noise_std_dev
        self.normalize_observations = normalize_observations
        self.normalize_returns = normalize_returns

    def instantiate(self, environment, device, settings) -> ReplayEnvRollerBase:
        return DequeReplayRollerOuNoise(
            environment=environment,
            device=device,
            batch_size=settings.batch_size,
            buffer_capacity=self.buffer_capacity,
            buffer_initial_size=self.buffer_initial_size,
            discount_factor=settings.discount_factor,
            noise_std_dev=self.noise_std_dev,
            normalize_observations=self.normalize_observations,
            normalize_returns=self.normalize_returns
        )


def create(buffer_capacity: int, buffer_initial_size: int, noise_std_dev: float,
           normalize_observations=False, normalize_returns=False):
    return DequeReplayRollerOuNoiseFactory(
        noise_std_dev=noise_std_dev,
        buffer_capacity=buffer_capacity,
        buffer_initial_size=buffer_initial_size,
        normalize_observations=normalize_observations,
        normalize_returns=normalize_returns
    )



import torch
import collections
import numpy as np

import vel.optimizers.rmsprop_tf as rprop_tf
import vel.rl.algo.policy_gradient.acer as acer
import vel.rl.models.q_policy_gradient_model as model
import vel.rl.models.backbone.nature_cnn as backbone

from vel.api import BatchInfo, EpochInfo, TrainingInfo


class AcerModelProxy:
    def __init__(self, ob_space, ac_space, nenvs, nsteps, nstack, num_procs,
                 ent_coef, q_coef, gamma, max_grad_norm, lr,
                 rprop_alpha, rprop_epsilon, total_timesteps, lrschedule,
                 c, trust_region, alpha, delta):
        self.observation_space = ob_space
        self.action_space = ac_space

        self.initial_state = None

        self.device = torch.device('cuda:0')

        self.model_factory = model.QPolicyGradientModelFactory(
            backbone.create(input_width=84, input_height=84, input_channels=4)
        )

        self.train_model = self.model_factory.instantiate(action_space=ac_space)

        self.train_model = self.train_model.to(self.device)
        self.train_model.reset_weights()

        self.optimizer = rprop_tf.RMSpropTF(
            self.train_model.parameters(),
            lr=lr,
            alpha=rprop_alpha,
            eps=rprop_epsilon
            # lr=7.0e-4,
            # alpha=0.99,
            # eps=1e-5,
        )

        self.real_acer = acer.AcerPolicyGradient(
            self.model_factory,
            trust_region=trust_region,
            entropy_coefficient=ent_coef,
            q_coefficient=q_coef,
            rho_cap=c,
            retrace_rho_cap=1.0,
            max_grad_norm=max_grad_norm,
            average_model_alpha=alpha, trust_region_delta=delta
        )

        self.num_envs = nenvs
        self.num_steps = nsteps

        settings = collections.namedtuple('settings', 'discount_factor number_of_steps action_space')
        s = settings(gamma, nsteps, ac_space)

        self.real_acer.initialize(s, self.train_model, s, self.device)

        self.training_info = TrainingInfo(start_epoch_idx=0)
        self.epoch_info = EpochInfo(
            training_info=self.training_info, global_epoch_idx=0, batches_per_epoch=1, optimizer=self.optimizer
        )

    def step(self, observations, S, M):
        device_observations = torch.from_numpy(observations).to(self.device)
        step_data = self.train_model.step(device_observations)

        return (
            step_data['actions'].cpu().numpy(),
            torch.exp(step_data['action_logits']).detach().cpu().numpy(),
            None
        )

    def train(self, obs, actions, rewards, dones, mus, states, masks, steps, on_policy=False):
        actions = actions.reshape(self.num_envs, self.num_steps).transpose(1, 0)
        rewards = rewards.reshape(self.num_envs, self.num_steps).transpose(1, 0)
        dones = dones.reshape(self.num_envs, self.num_steps).transpose(1, 0)
        obs = obs.reshape(self.num_envs, (self.num_steps + 1), 84, 84, 4).transpose(1, 0, 2, 3, 4)
        mus = mus.reshape(self.num_envs, self.num_steps, self.action_space.n).transpose(1, 0, 2)

        logits = np.log(mus)

        # Strip from last observations
        observations = obs[:20]

        final_values = self.train_model.value(torch.from_numpy(obs[-1]).to(self.device)).detach()

        rollout = {
            'actions': torch.from_numpy(actions).long().reshape(self.num_envs * self.num_steps).to(self.device),
            'rewards': torch.from_numpy(rewards).reshape(self.num_envs * self.num_steps).to(self.device),
            'dones': torch.from_numpy(dones.astype(np.uint8)).reshape(self.num_envs * self.num_steps).to(self.device),
            'observations': torch.from_numpy(observations).reshape(self.num_envs * self.num_steps, 84, 84, 4).to(self.device),
            'final_values': final_values,
            'action_logits': torch.from_numpy(logits).reshape(self.num_envs * self.num_steps, self.action_space.n).to(self.device)
        }

        batch_info = BatchInfo(self.epoch_info, batch_number=0)

        batch_info['sub_batch_data'] = []

        self.real_acer.optimizer_step(batch_info, self.device, self.train_model, rollout)

        return tuple(zip(*batch_info['sub_batch_data'][0].items()))



from vel.api import TrainingInfo


class BaseMetric:
    """ Base class for all the metrics """

    def __init__(self, name):
        self.name = name

    def calculate(self, batch_info):
        """ Calculate value of a metric based on supplied data """
        raise NotImplementedError

    def reset(self):
        """ Reset value of a metric """
        raise NotImplementedError

    def value(self):
        """ Return current value for the metric """
        raise NotImplementedError

    def write_state_dict(self, training_info: TrainingInfo, hidden_state_dict: dict) -> None:
        """ Potentially store some metric state to the checkpoint """
        pass

    def load_state_dict(self, training_info: TrainingInfo, hidden_state_dict: dict) -> None:
        """ Potentially load some metric state from the checkpoint """
        pass



from vel.openai.baselines.common.vec_env import VecEnvWrapper
from vel.openai.baselines.common.running_mean_std import RunningMeanStd
import numpy as np

class VecNormalize(VecEnvWrapper):
    """
    Vectorized environment base class
    """
    def __init__(self, venv, ob=True, ret=True, clipob=10., cliprew=10., gamma=0.99, epsilon=1e-8):
        VecEnvWrapper.__init__(self, venv)
        self.ob_rms = RunningMeanStd(shape=self.observation_space.shape) if ob else None
        self.ret_rms = RunningMeanStd(shape=()) if ret else None
        self.clipob = clipob
        self.cliprew = cliprew
        self.ret = np.zeros(self.num_envs)
        self.gamma = gamma
        self.epsilon = epsilon

    def step_wait(self):
        """
        Apply sequence of actions to sequence of environments
        actions -> (observations, rewards, news)

        where 'news' is a boolean vector indicating whether each element is new.
        """
        obs, rews, news, infos = self.venv.step_wait()
        self.ret = self.ret * self.gamma + rews
        obs = self._obfilt(obs)
        if self.ret_rms:
            self.ret_rms.update(self.ret)
            rews = np.clip(rews / np.sqrt(self.ret_rms.var + self.epsilon), -self.cliprew, self.cliprew)
        return obs, rews, news, infos

    def _obfilt(self, obs):
        if self.ob_rms:
            self.ob_rms.update(obs)
            obs = np.clip((obs - self.ob_rms.mean) / np.sqrt(self.ob_rms.var + self.epsilon), -self.clipob, self.clipob)
            return obs.astype(np.float32)
        else:
            return obs

    def reset(self):
        """
        Reset all environments
        """
        obs = self.venv.reset()
        return self._obfilt(obs)



import collections

import numpy as np
import torch

from vel.api import BatchInfo
from vel.api.metrics import BaseMetric, AveragingMetric, ValueMetric


class FramesMetric(ValueMetric):
    """ Count the frames """
    def __init__(self, name="frames"):
        super().__init__(name)

    def _value_function(self, batch_info: BatchInfo):
        return batch_info.training_info['frames']


class FPSMetric(ValueMetric):
    """ Metric calculating FPS values """
    def __init__(self, name='fps'):
        super().__init__(name)

    def _value_function(self, batch_info):
        frames = batch_info.training_info['frames']
        seconds = batch_info.training_info['time']

        fps = int(frames/seconds)

        return fps


class EpisodeRewardMetric(BaseMetric):
    def __init__(self, name):
        super().__init__(name)
        self.buffer = collections.deque(maxlen=100)

    def calculate(self, batch_info):
        """ Calculate value of a metric based on supplied data """
        self.buffer.extend(batch_info['episode_infos'])

    def reset(self):
        """ Reset value of a metric """
        # Because it's a queue no need for reset..
        pass

    def value(self):
        """ Return current value for the metric """
        if self.buffer:
            return np.mean([ep['r'] for ep in self.buffer])
        else:
            return 0.0


class EpisodeRewardMetricQuantile(BaseMetric):
    def __init__(self, name, quantile, buf_size=100):
        super().__init__(name)
        self.buffer = collections.deque(maxlen=buf_size)
        self.quantile = quantile

    def calculate(self, batch_info):
        """ Calculate value of a metric based on supplied data """
        self.buffer.extend(ep['r'] for ep in batch_info['episode_infos'])

    def reset(self):
        """ Reset value of a metric """
        # Because it's a queue no need for reset..
        pass

    def value(self):
        """ Return current value for the metric """
        if self.buffer:
            return np.quantile(self.buffer, self.quantile)
        else:
            return 0.0


class EpisodeLengthMetric(BaseMetric):
    def __init__(self, name):
        super().__init__(name)
        self.buffer = collections.deque(maxlen=100)

    def calculate(self, batch_info):
        """ Calculate value of a metric based on supplied data """
        self.buffer.extend(batch_info['episode_infos'])

    def reset(self):
        """ Reset value of a metric """
        # Because it's a queue no need for reset..
        pass

    def value(self):
        """ Return current value for the metric """
        if self.buffer:
            return np.mean([ep['l'] for ep in self.buffer])
        else:
            return 0


class ExplainedVariance(AveragingMetric):
    """ How much value do rewards explain """
    def __init__(self):
        super().__init__("explained_variance")

    def _value_function(self, batch_info):
        values = batch_info['values']
        rewards = batch_info['rewards']

        explained_variance = 1 - torch.var(rewards - values) / torch.var(rewards)
        return explained_variance.item()



import torch.optim

from vel.api.base import OptimizerFactory, Model


class AdadeltaFactory(OptimizerFactory):
    """ Adadelta optimizer factory """

    def __init__(self, lr=1.0, rho=0.9, eps=1e-6, weight_decay=0):
        self.lr = lr
        self.rho = rho
        self.eps = eps
        self.weight_decay = weight_decay

    def instantiate(self, model: Model) -> torch.optim.Adadelta:
        return torch.optim.Adadelta(
            filter(lambda p: p.requires_grad, model.parameters()),
            lr=self.lr, rho=self.rho, eps=self.eps, weight_decay=self.weight_decay
        )


def create():
    """ Return an ADADELTA optimizer """
    return AdadeltaFactory()



"""
Code based on
https://github.com/fastai/fastai/blob/master/fastai/model.py
"""
import torch.nn as nn

import collections
import itertools as it


def is_listy(x):
    return isinstance(x, (list, tuple))


def model_children(module):
    return module if isinstance(module, (list, tuple)) else list(module.children())


def apply_leaf(module, f):
    c = model_children(module)

    if isinstance(module, nn.Module):
        f(module)

    if len(c) > 0:
        for l in c:
            apply_leaf(l, f)


def set_train_mode(module):
    # Only fix ones which we don't want to "train"
    if hasattr(module, 'running_mean') and (getattr(module, 'bn_freeze', False) or not getattr(module, 'trainable', True)):
        module.eval()
    elif getattr(module, 'drop_freeze', False) and hasattr(module, 'p') and ('drop' in type(module).__name__.lower()):
        module.eval()


def set_trainable_attr(module, trainable):
    module.trainable = trainable


def set_requires_gradient(module, trainable):
    for p in module.parameters():
        p.requires_grad = trainable


def freeze_layer(module):
    apply_leaf(module, lambda x: set_trainable_attr(x, trainable=False))
    set_requires_gradient(module, trainable=False)


def unfreeze_layer(module):
    apply_leaf(module, lambda x: set_trainable_attr(x, trainable=True))
    set_requires_gradient(module, trainable=True)


def trainable_params_(m):
    """ Returns a list of trainable parameters in the model m. (i.e., those that require gradients.) """
    if isinstance(m, collections.Sequence):
        return [p for p in m if p.requires_grad]
    else:
        return [p for p in m.parameters() if p.requires_grad]


def chain_params(p):
    if is_listy(p):
        return list(it.chain(*[trainable_params_(o) for o in p]))
    return trainable_params_(p)


def to_parameter_groups(layer_groups):
    """ Convert from list of layer groups into list of parameter settings for an optimizer """
    return [{'params': chain_params(x)} for x in layer_groups]



import torch
import typing

from vel.api import Learner, ModelConfig, EpochInfo, TrainingInfo
from vel.api.base import OptimizerFactory, SchedulerFactory, Callback, Source, Storage, ModelFactory
from vel.callbacks.time_tracker import TimeTracker


class SimpleTrainCommand:
    """ Very simple training command - just run the supplied generators """

    def __init__(self, epochs: int, model_config: ModelConfig, model_factory: ModelFactory,
                 optimizer_factory: OptimizerFactory, scheduler_factory: typing.Optional[SchedulerFactory],
                 source: Source, storage: Storage, callbacks: typing.Optional[typing.List[Callback]],
                 max_grad_norm: typing.Optional[float]):
        self.epochs = epochs
        self.model_config = model_config
        self.model_factory = model_factory

        self.optimizer_factory = optimizer_factory
        self.scheduler_factory = scheduler_factory

        self.source = source
        self.storage = storage
        self.callbacks = callbacks if callbacks is not None else []
        self.max_grad_norm = max_grad_norm

    def run(self):
        """ Run the command with supplied configuration """
        device = torch.device(self.model_config.device)
        learner = Learner(device, self.model_factory.instantiate(), self.max_grad_norm)
        optimizer = self.optimizer_factory.instantiate(learner.model)

        # All callbacks used for learning
        callbacks = self.gather_callbacks(optimizer)

        # Metrics to track through this training
        metrics = learner.metrics()

        # Check if training was already started and potentially continue where we left off
        training_info = self.resume_training(learner, callbacks, metrics)

        training_info.on_train_begin()

        if training_info.optimizer_initial_state:
            optimizer.load_state_dict(training_info.optimizer_initial_state)

        for global_epoch_idx in range(training_info.start_epoch_idx + 1, self.epochs + 1):
            epoch_info = EpochInfo(
                training_info=training_info,
                global_epoch_idx=global_epoch_idx,
                batches_per_epoch=self.source.train_iterations_per_epoch(),
                optimizer=optimizer
            )

            # Execute learning
            learner.run_epoch(epoch_info, self.source)

            self.storage.checkpoint(epoch_info, learner.model)

        training_info.on_train_end()

        return training_info

    def gather_callbacks(self, optimizer) -> list:
        """ Gather all the callbacks to be used in this training run """
        callbacks = [TimeTracker()]

        if self.scheduler_factory is not None:
            callbacks.append(self.scheduler_factory.instantiate(optimizer))

        callbacks.extend(self.callbacks)
        callbacks.extend(self.storage.streaming_callbacks())

        return callbacks

    def resume_training(self, learner, callbacks, metrics) -> TrainingInfo:
        """ Possibly resume training from a saved state from the storage """
        if self.model_config.continue_training:
            start_epoch = self.storage.last_epoch_idx()
        else:
            start_epoch = 0

        training_info = TrainingInfo(
            start_epoch_idx=start_epoch,
            run_name=self.model_config.run_name,
            metrics=metrics,
            callbacks=callbacks
        )

        if start_epoch == 0:
            self.storage.reset(self.model_config.render_configuration())
            training_info.initialize()
            learner.initialize_training(training_info)
        else:
            self.storage.resume(training_info, learner.model)

        return training_info


def create(model_config, epochs, optimizer, model, source, storage, scheduler=None, callbacks=None, max_grad_norm=None):
    """ Simply train the model """
    return SimpleTrainCommand(
        epochs=epochs,
        model_config=model_config,
        model_factory=model,
        optimizer_factory=optimizer,
        scheduler_factory=scheduler,
        source=source,
        storage=storage,
        callbacks=callbacks,
        max_grad_norm=max_grad_norm
    )



"""
Code based on
https://github.com/pytorch/vision/blob/master/torchvision/models/resnet.py
"""

import torch.nn as nn
import torch.nn.functional as F

from vel.api.base import SupervisedModel, ModelFactory
from vel.modules.resnet_v1 import Bottleneck, BasicBlock


class ResNetV1(SupervisedModel):
    """ A ResNet V1 model as defined in the literature """

    def __init__(self, block, layers, inplanes, divisor=4, img_channels=3, num_classes=1000):
        super().__init__()

        self.num_classess = num_classes
        self.inplanes = inplanes
        self.divisor = divisor

        self.pre_conv = nn.Conv2d(img_channels, inplanes, kernel_size=(3, 3), padding=1, bias=False)
        self.pre_bn = nn.BatchNorm2d(inplanes)

        self.layer1 = self._make_layer(block, inplanes,     inplanes,     layers[0], stride=1)
        self.layer2 = self._make_layer(block, inplanes,     inplanes * 2, layers[1], stride=2)
        self.layer3 = self._make_layer(block, inplanes * 2, inplanes * 4, layers[2], stride=2)

        self.avgpool = nn.AdaptiveAvgPool2d(1)
        self.fc = nn.Linear(inplanes * 4, num_classes)

        # Weight initialization
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode='fan_out', nonlinearity='relu')
            elif isinstance(m, nn.Linear):
                nn.init.kaiming_normal_(m.weight, mode='fan_out', nonlinearity='relu')
                nn.init.constant_(m.bias, 0.0)
            elif isinstance(m, nn.BatchNorm2d):
                nn.init.constant_(m.weight, 1.0)
                nn.init.constant_(m.bias, 0.0)

    def _make_layer(self, block, in_channels, out_channels, blocks, stride=1):
        layers = []

        # First block is the one doing the rescaling
        layers.append(block(in_channels, out_channels, stride, divisor=self.divisor))

        for i in range(1, blocks):
            layers.append(block(out_channels, out_channels, divisor=self.divisor))

        return nn.Sequential(*layers)

    def forward(self, x):
        x = self.pre_conv(x)
        x = self.pre_bn(x)
        x = F.relu(x)

        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)

        x = self.avgpool(x)
        x = x.view(x.size(0), -1)
        x = self.fc(x)
        x = F.log_softmax(x, dim=1)

        return x

    def loss_value(self, x_data, y_true, y_pred):
        """ Calculate value of the loss function """
        return F.nll_loss(y_pred, y_true)

    def metrics(self):
        """ Set of metrics for this model """
        from vel.metrics.loss_metric import Loss
        from vel.metrics.accuracy import Accuracy
        return [Loss(), Accuracy()]


def create(blocks, mode='basic', inplanes=16, divisor=4, num_classes=1000):
    """
    Create a ResNetV1 model
    """
    block_dict = {
        'basic': BasicBlock,
        'bottleneck': Bottleneck
    }

    def instantiate(**_):
        return ResNetV1(block_dict[mode], blocks, inplanes=inplanes, divisor=divisor, num_classes=num_classes)

    return ModelFactory.generic(instantiate)



from vel.api.metrics.averaging_metric import AveragingMetric


class Loss(AveragingMetric):
    """ Just a loss function """
    def __init__(self):
        super().__init__("loss")

    def _value_function(self, batch_info):
        """ Just forward a value of the loss"""
        return batch_info['loss'].item()



"""
Code based on:
https://github.com/fastai/fastai/blob/master/fastai/transforms.py
"""
import cv2
import collections
import random

import vel.api.data as data


class RandomScale(data.Augmentation):
    """ Scales the image so that the smallest axis is of 'size' times a random number between 1.0 and max_zoom. """
    def __init__(self, size, max_zoom, p=0.75, mode='x', tags=None):
        super().__init__(mode, tags)
        self.size = size
        self.max_zoom = max_zoom
        self.p = p

    def __call__(self, x_data):
        if random.random() < self.p:
            # Yes, do it
            min_z = 1.
            max_z = self.max_zoom
            if isinstance(self.max_zoom, collections.Iterable):
                min_z, max_z = self.max_zoom

            mult = random.uniform(min_z, max_z)
        else:
            # No, don't do it
            mult = 1.0

        return data.scale_min(x_data, int(self.size * mult), cv2.INTER_AREA)


def create(size, max_zoom, p=0.75, mode='x', tags=None):
    return RandomScale(size, max_zoom, p, mode, tags)



import gym

from vel.api.base import LinearBackboneModel, Model, ModelFactory
from vel.rl.api import Rollout, Evaluator
from vel.rl.modules.dueling_q_head import DuelingQHead
from vel.rl.models.q_model import QModelEvaluator


class QDuelingModel(Model):
    """ Wraps a backbone model into API we need for Deep Q-Learning """

    def __init__(self, backbone: LinearBackboneModel, action_space: gym.Space):
        super().__init__()

        self.backbone = backbone
        self.q_head = DuelingQHead(input_dim=backbone.output_dim, action_space=action_space)

    def forward(self, observations):
        """ Model forward pass """
        advantage_features, value_features = self.backbone(observations)
        q_values = self.q_head(advantage_features, value_features)

        return q_values

    def reset_weights(self):
        """ Initialize weights to reasonable defaults """
        self.backbone.reset_weights()
        self.q_head.reset_weights()

    def step(self, observations):
        """ Sample action from an action space for given state """
        q_values = self(observations)

        return {
            'actions': self.q_head.sample(q_values),
            'values': q_values
        }

    def evaluate(self, rollout: Rollout) -> Evaluator:
        """ Evaluate model on a rollout """
        return QModelEvaluator(self, rollout)


class QDuelingModelFactory(ModelFactory):
    """ Factory  class for policy gradient models """
    def __init__(self, backbone: ModelFactory):
        self.backbone = backbone

    def instantiate(self, **extra_args):
        """ Instantiate the model """
        backbone = self.backbone.instantiate(**extra_args)
        return QDuelingModel(backbone, extra_args['action_space'])


def create(backbone: ModelFactory):
    """ DQN model factory """
    return QDuelingModelFactory(backbone=backbone)



"""
Code based on:
https://github.com/fastai/fastai/blob/master/fastai/transforms.py
"""
import cv2
import random

import vel.api.data as data


class RandomRotate(data.Augmentation):
    """ Rotate image randomly by an angle between (-deg, +deg) """
    def __init__(self, deg, p=0.75, mode='x', tags=None):
        super().__init__(mode, tags)
        self.deg = deg
        self.p = p

    def __call__(self, x_data):
        if random.random() < self.p:
            random_degree = random.uniform(-self.deg, self.deg)
            return data.rotate_img(x_data, random_degree, mode=cv2.BORDER_REFLECT)
        else:
            # No, don't do it
            return x_data


def create(deg, p=0.75, mode='x', tags=None):
    return RandomRotate(deg, p, mode, tags)



"""
Code based on
https://github.com/pytorch/vision/blob/master/torchvision/models/resnet.py
"""

import torch.nn as nn
import torch.nn.functional as F


def conv3x3(in_channels, out_channels, stride=1):
    """
    3x3 convolution with padding.
    Original code has had bias turned off, because Batch Norm would remove the bias either way
    """
    return nn.Conv2d(in_channels, out_channels, kernel_size=3, stride=stride, padding=1, bias=False)


class BasicBlock(nn.Module):
    """
    Single residual block consisting of two convolutional layers and a nonlinearity between them
    """
    def __init__(self, in_channels, out_channels, stride=1, divisor=None):
        super().__init__()

        self.stride = stride
        self.divisor = divisor

        if stride != 1 or in_channels != out_channels:
            self.shortcut = nn.Conv2d(in_channels, out_channels, kernel_size=1, stride=stride, bias=False)
        else:
            self.shortcut = None

        self.bn1 = nn.BatchNorm2d(in_channels)
        self.conv1 = conv3x3(in_channels, out_channels, stride)

        self.bn2 = nn.BatchNorm2d(out_channels)
        self.conv2 = conv3x3(out_channels, out_channels)

    def forward(self, x):
        out = self.bn1(x)
        out = F.relu(out)

        # If we need to rescale the "main path"
        if self.shortcut:
            residual = self.shortcut(out)
        else:
            residual = x

        out = self.conv1(out)

        out = self.bn2(out)
        out = F.relu(out)
        out = self.conv2(out)

        out += residual

        return out


class Bottleneck(nn.Module):
    """
    A 'bottleneck' residual block consisting of three convolutional layers, where the first one is a downsampler,
    then we have a 3x3 followed by an upsampler.
    """
    def __init__(self, in_channels, out_channels, stride=1, divisor=4):
        super().__init__()

        self.stride = stride
        self.divisor = divisor

        if stride != 1 or in_channels != out_channels:
            self.shortcut = nn.Sequential(
                nn.Conv2d(in_channels, out_channels, kernel_size=1, stride=stride, bias=False)
            )
        else:
            self.shortcut = None

        self.bottleneck_channels = out_channels // divisor

        self.bn1 = nn.BatchNorm2d(in_channels)
        self.conv1 = nn.Conv2d(in_channels, self.bottleneck_channels, kernel_size=1, bias=False)

        self.bn2 = nn.BatchNorm2d(self.bottleneck_channels)
        self.conv2 = nn.Conv2d(self.bottleneck_channels, self.bottleneck_channels, kernel_size=3, stride=stride, padding=1, bias=False)

        self.bn3 = nn.BatchNorm2d(self.bottleneck_channels)
        self.conv3 = nn.Conv2d(self.bottleneck_channels, out_channels, kernel_size=1, bias=False)

    def forward(self, x):
        out = F.relu(self.bn1(x))
        # shortcut = self.shortcut(out) if hasattr(self, 'shortcut') else x
        # If we need to rescale the "main path"

        if self.shortcut:
            residual = self.shortcut(out)
        else:
            residual = x

        out = self.conv1(out)
        out = self.conv2(F.relu(self.bn2(out)))
        out = self.conv3(F.relu(self.bn3(out)))

        out += residual

        return out



import random

import vel.api.data as data


class RandomLighting(data.Augmentation):
    """ Apply a horizontal flip randomly to input images """

    def __init__(self, b, c, mode='x', tags=None):
        super().__init__(mode, tags)
        self.b, self.c = b, c

    def __call__(self, img):
        """ Adjust lighting """
        rand_b = random.uniform(-self.b, self.b)
        rand_c = random.uniform(-self.c, self.c)
        rand_c = -1/(rand_c-1) if rand_c<0 else rand_c+1
        return data.lighting(img, rand_b, rand_c)

    def __repr__(self):
        return self.__class__.__name__ + '(p={})'.format(self.p)


def create(b, c, mode='x', tags=None):
    return RandomLighting(b, c, mode, tags)



import visdom


from vel.util.visdom import visdom_push_metrics, VisdomSettings


class VisdomCommand:
    """ Send metrics from database into VISDOM """
    def __init__(self, model_config, storage, visdom_settings: VisdomSettings):
        self.model_config = model_config
        self.storage = storage
        self.vis = visdom.Visdom(
            server=visdom_settings.server,
            endpoint=visdom_settings.endpoint,
            port=visdom_settings.port,
            env=self.model_config.run_name.replace('/', '_')
        )

    def run(self):
        metrics = self.storage.get_metrics_frame().drop('run_name', axis=1)
        visdom_push_metrics(self.vis, metrics)


def create(model_config, storage, visdom_settings):
    return VisdomCommand(model_config, storage, VisdomSettings(**visdom_settings))



import numpy as np
import gym.spaces as spaces

import torch
import torch.nn as nn
import torch.nn.init as init


class DeterministicActionHead(nn.Module):
    """
    Network head for action determination. Returns deterministic action depending on the inputs
    """

    def __init__(self, input_dim, action_space):
        super().__init__()

        self.action_space = action_space

        assert isinstance(action_space, spaces.Box)
        assert len(action_space.shape) == 1

        assert (np.abs(action_space.low) == action_space.high).all()  # we assume symmetric actions.
        self.register_buffer('max_action', torch.from_numpy(action_space.high))

        self.linear_layer = nn.Linear(input_dim, action_space.shape[0])

    def forward(self, input_data):
        return torch.tanh(self.linear_layer(input_data)) * self.max_action

    def sample(self, params, **_):
        """ Sample from a probability space of all actions """
        return {
            'actions': self(params)
        }

    def reset_weights(self):
        """ Initialize weights to sane defaults """
        init.orthogonal_(self.linear_layer.weight, gain=0.01)
        init.constant_(self.linear_layer.bias, 0.0)



import numpy as np
import pandas as pd
import torch

from vel.api import ModelConfig, TrainingInfo
from vel.api.base import Storage, ModelFactory
from vel.rl.api.base import EnvFactory
from vel.openai.baselines.common.atari_wrappers import FrameStack


class EvaluateEnvCommand:
    """ Record environment playthrough as a game  """
    def __init__(self, model_config: ModelConfig, env_factory: EnvFactory, model_factory: ModelFactory,
                 storage: Storage, takes: int, frame_history: int, sample_args: dict = None):
        self.model_config = model_config
        self.model_factory = model_factory
        self.env_factory = env_factory
        self.storage = storage
        self.takes = takes
        self.frame_history = frame_history
        self.sample_args = sample_args if sample_args is not None else {}

    def run(self):
        device = torch.device(self.model_config.device)

        env = FrameStack(self.env_factory.instantiate(preset='raw'), self.frame_history)
        model = self.model_factory.instantiate(action_space=env.action_space).to(device)

        training_info = TrainingInfo(start_epoch_idx=self.storage.last_epoch_idx(), run_name=self.model_config.run_name)
        self.storage.resume(training_info, model)

        model.eval()

        rewards = []
        lengths = []

        for i in range(self.takes):
            result = self.record_take(model, env, device, takenumber=i+1)
            rewards.append(result['r'])
            lengths.append(result['l'])

        print(pd.DataFrame({'lengths': lengths, 'rewards': rewards}).describe())

    @torch.no_grad()
    def record_take(self, model, env_instance, device, takenumber):
        frames = []

        observation = env_instance.reset()

        frames.append(env_instance.render('rgb_array'))

        print("Evaluating environment...")

        while True:
            observation_array = np.expand_dims(np.array(observation), axis=0)
            observation_tensor = torch.from_numpy(observation_array).to(device)
            actions = model.step(observation_tensor, **self.sample_args)['actions']

            observation, reward, done, epinfo = env_instance.step(actions.item())

            frames.append(env_instance.render('rgb_array'))

            if 'episode' in epinfo:
                # End of an episode
                return epinfo['episode']


def create(model_config, model, env, storage, takes, frame_history, sample_args=None):
    return EvaluateEnvCommand(
        model_config=model_config,
        model_factory=model,
        env_factory=env,
        storage=storage,
        frame_history=frame_history,
        takes=takes,
        sample_args=sample_args
    )



import torch.optim.lr_scheduler as scheduler


# class MultiStepScheduler:
#     def __init__(self, optimizer, milestones, gamma, last_epoch):
#         self.scheduler = scheduler.MultiStepLR(optimizer, milestones, gamma, last_epoch=last_epoch)
#
#     def get_lr(self):
#         return self.scheduler.get_lr()
#
#     def pre_epoch_step(self, epoch_idx):
#         self.scheduler.step()
#
#     def post_epoch_step(self, epoch_idx, metrics):
#         pass
#
#
# def create(milestones, gamma=0.1):
#     """ Create a multi-step scheduler """
#     def scheduler_fn(optimizer, last_epoch=-1):
#         return MultiStepScheduler(optimizer, milestones, gamma, last_epoch=last_epoch)
#
#     return scheduler_fn



import numpy as np
import torch
import torch.autograd as autograd
import torch.nn.functional as F
import torch.nn.utils

from vel.api.metrics.averaging_metric import AveragingNamedMetric
from vel.math.functions import explained_variance
from vel.rl.api.base import AlgoBase


def p2v(params):
    """ Parameters to vector - shorthand utility version """
    return torch.nn.utils.parameters_to_vector(params)


def v2p(vector, params):
    """ Vector to parameters - shorthand utility version """
    return torch.nn.utils.vector_to_parameters(vector, params)


def conjugate_gradient_method(matrix_vector_operator, loss_gradient, nsteps, rdotr_tol=1e-10):
    """ Conjugate gradient algorithm """
    x = torch.zeros_like(loss_gradient)

    r = loss_gradient.clone()
    p = loss_gradient.clone()

    rdotr = torch.dot(r, r)

    for i in range(nsteps):
        Avp = matrix_vector_operator(p)
        alpha = rdotr / torch.dot(p, Avp)

        x += alpha * p
        r -= alpha * Avp

        new_rdotr = torch.dot(r, r)
        betta = new_rdotr / rdotr
        p = r + betta * p
        rdotr = new_rdotr

        if rdotr < rdotr_tol:
            break

    return x


class TrpoPolicyGradient(AlgoBase):
    """ Trust Region Policy Optimization - https://arxiv.org/abs/1502.05477 """

    def __init__(self, max_kl, cg_iters, line_search_iters, cg_damping, entropy_coef, vf_iters,
                 improvement_acceptance_ratio, max_grad_norm):
        self.mak_kl = max_kl
        self.cg_iters = cg_iters
        self.line_search_iters = line_search_iters
        self.cg_damping = cg_damping
        self.entropy_coef = entropy_coef
        self.vf_iters = vf_iters
        self.improvement_acceptance_ratio = improvement_acceptance_ratio
        self.max_grad_norm = max_grad_norm

    def optimizer_step(self, batch_info, device, model, rollout):
        """ Single optimization step for a model """
        rollout = rollout.to_transitions()

        # This algorithm makes quote strong assumptions about how does the model look
        # so it does not make that much sense to switch to the evaluator interface
        # As it would be more of a problem than actual benefit

        observations = rollout.batch_tensor('observations')
        returns = rollout.batch_tensor('estimated_returns')

        # Evaluate model on the observations
        policy_params = model.policy(observations)
        policy_entropy = torch.mean(model.entropy(policy_params))

        policy_loss = self.calc_policy_loss(model, policy_params, policy_entropy, rollout)
        policy_grad = p2v(autograd.grad(policy_loss, model.policy_parameters(), retain_graph=True)).detach()

        # Calculate gradient of KL divergence of model with fixed version of itself
        # Value of kl_divergence will be 0, but what we need is the gradient, actually the 2nd derivarive
        kl_divergence = torch.mean(model.kl_divergence(policy_params.detach(), policy_params))
        kl_divergence_gradient = p2v(torch.autograd.grad(kl_divergence, model.policy_parameters(), create_graph=True))

        step_direction = conjugate_gradient_method(
            matrix_vector_operator=lambda x: self.fisher_vector_product(x, kl_divergence_gradient, model),
            # Because we want to decrease the loss, we want to go into the direction of -gradient
            loss_gradient=-policy_grad,
            nsteps=self.cg_iters
        )

        shs = 0.5 * step_direction @ self.fisher_vector_product(step_direction, kl_divergence_gradient, model)
        lm = torch.sqrt(shs / self.mak_kl)
        full_step = step_direction / lm

        # Because we want to decrease the loss, we want to go into the direction of -gradient
        expected_improvement = (-policy_grad) @ full_step
        original_parameter_vec = p2v(model.policy_parameters()).detach_()

        policy_optimization_success, ratio, policy_loss_improvement, new_policy_loss, kl_divergence_step = self.line_search(
            model, rollout, policy_loss, policy_params, original_parameter_vec, full_step, expected_improvement
        )

        gradient_norms = []

        for i in range(self.vf_iters):
            batch_info.optimizer.zero_grad()
            value_loss = self.value_loss(model, observations, returns)

            value_loss.backward()

            # Gradient clipping
            if self.max_grad_norm is not None:
                grad_norm = torch.nn.utils.clip_grad_norm_(
                    filter(lambda p: p.requires_grad, model.parameters()),
                    max_norm=self.max_grad_norm
                )

                gradient_norms.append(grad_norm)

            batch_info.optimizer.step(closure=None)

        if gradient_norms:
            gradient_norm = np.mean(gradient_norms)
        else:
            gradient_norm = 0.0

        # noinspection PyUnboundLocalVariable
        return {
            'new_policy_loss': new_policy_loss.item(),
            'policy_entropy': policy_entropy.item(),
            'value_loss': value_loss.item(),
            'policy_optimization_success': float(policy_optimization_success),
            'policy_improvement_ratio': ratio.item(),
            'kl_divergence_step': kl_divergence_step.item(),
            'policy_loss_improvement': policy_loss_improvement.item(),
            'grad_norm': gradient_norm,
            'advantage_norm': torch.norm(rollout.batch_tensor('estimated_advantages')).item(),
            'explained_variance': explained_variance(returns, rollout.batch_tensor('estimated_values'))
        }

    def line_search(self, model, rollout, original_policy_loss, original_policy_params, original_parameter_vec,
                    full_step, expected_improvement_full):
        """ Find the right stepsize to make sure policy improves """
        current_parameter_vec = original_parameter_vec.clone()

        for idx in range(self.line_search_iters):
            stepsize = 0.5 ** idx

            new_parameter_vec = current_parameter_vec + stepsize * full_step

            # Update model parameters
            v2p(new_parameter_vec, model.policy_parameters())

            # Calculate new loss
            with torch.no_grad():
                policy_params = model.policy(rollout.batch_tensor('observations'))
                policy_entropy = torch.mean(model.entropy(policy_params))
                kl_divergence = torch.mean(model.kl_divergence(original_policy_params, policy_params))

                new_loss = self.calc_policy_loss(model, policy_params, policy_entropy, rollout)

                actual_improvement = original_policy_loss - new_loss
                expected_improvement = expected_improvement_full * stepsize

                ratio = actual_improvement / expected_improvement

            if kl_divergence.item() > self.mak_kl * 1.5:
                # KL divergence bound exceeded
                continue
            elif ratio < expected_improvement:
                # Not enough loss improvement
                continue
            else:
                # Optimization successful
                return True, ratio, actual_improvement, new_loss, kl_divergence

        # Optimization failed, revert to initial parameters
        v2p(original_parameter_vec, model.policy_parameters())
        return False, torch.tensor(0.0), torch.tensor(0.0), torch.tensor(0.0), torch.tensor(0.0)

    def fisher_vector_product(self, vector, kl_divergence_gradient, model):
        """ Calculate product Hessian @ vector """
        assert not vector.requires_grad, "Vector must not propagate gradient"
        dot_product = vector @ kl_divergence_gradient

        # at least one dimension spans across two contiguous subspaces
        double_gradient = torch.autograd.grad(dot_product, model.policy_parameters(), retain_graph=True)
        fvp = p2v(x.contiguous() for x in double_gradient)

        return fvp + vector * self.cg_damping

    def value_loss(self, model, observations, discounted_rewards):
        """ Loss of value estimator """
        value_outputs = model.value(observations)
        value_loss = 0.5 * F.mse_loss(value_outputs, discounted_rewards)
        return value_loss

    def calc_policy_loss(self, model, policy_params, policy_entropy, rollout):
        """
        Policy gradient loss - calculate from probability distribution

        Calculate surrogate loss - advantage * policy_probability / fixed_initial_policy_probability

        Because we operate with logarithm of -probability (neglogp) we do
        - advantage * exp(fixed_neglogps - model_neglogps)
        """
        actions = rollout.batch_tensor('actions')
        advantages = rollout.batch_tensor('estimated_advantages')
        fixed_logprobs = rollout.batch_tensor('action:logprobs')

        model_logprobs = model.logprob(actions, policy_params)

        # Normalize advantages
        advantages = (advantages - advantages.mean()) / (advantages.std() + 1e-8)

        # We put - in front because we want to maximize the surrogate objective
        policy_loss = -advantages * torch.exp(model_logprobs - fixed_logprobs)

        return policy_loss.mean() - policy_entropy * self.entropy_coef

    def metrics(self) -> list:
        """ List of metrics to track for this learning process """
        return [
            AveragingNamedMetric("new_policy_loss"),
            AveragingNamedMetric("policy_entropy"),
            AveragingNamedMetric("value_loss"),
            AveragingNamedMetric("policy_optimization_success"),
            AveragingNamedMetric("policy_improvement_ratio"),
            AveragingNamedMetric("kl_divergence_step"),
            AveragingNamedMetric("policy_loss_improvement"),
            AveragingNamedMetric("grad_norm"),
            AveragingNamedMetric("advantage_norm"),
            AveragingNamedMetric("explained_variance")
        ]


def create(max_kl, cg_iters, line_search_iters, cg_damping, entropy_coef, vf_iters, improvement_acceptance_ratio=0.1,
           max_grad_norm=0.5):
    return TrpoPolicyGradient(
        max_kl, int(cg_iters), int(line_search_iters), cg_damping, entropy_coef, vf_iters, improvement_acceptance_ratio,
        max_grad_norm=max_grad_norm
    )



from vel.api.base import Model

from torch.optim import Optimizer


class OptimizerFactory:
    """ Base class for optimizer factories """

    def instantiate(self, model: Model) -> Optimizer:
        raise NotImplementedError



import typing

from .checkpoint_strategy import CheckpointStrategy
from vel.util.better import better


class ClassicCheckpointStrategy(CheckpointStrategy):
    """ Classic checkpoint strategy """
    def __init__(self, checkpoint_frequency=0, metric=None, metric_mode='min', store_best=False):
        self.checkpoint_frequency = checkpoint_frequency
        self.metric = metric or 'val:loss'
        self.metric_mode = metric_mode
        self.store_best = store_best

        # TODO(jerry) initialize these values from hidden state
        self._current_best_metric_value = None
        self._current_best_checkpoint_idx = None

    def should_delete_previous_checkpoint(self, epoch_idx) -> bool:
        prev_epoch_idx = epoch_idx - 1

        if self.checkpoint_frequency > 0 and prev_epoch_idx % self.checkpoint_frequency == 0:
            return False
        else:
            return True

    def should_store_best_checkpoint(self, epoch_idx, metrics) -> bool:
        """ Should we store current checkpoint as the best """
        if not self.store_best:
            return False

        metric = metrics[self.metric]

        if better(self._current_best_metric_value, metric, self.metric_mode):
            self._current_best_metric_value = metric
            return True

        return False

    def store_best_checkpoint_idx(self, epoch_idx) -> None:
        """ Should we store current checkpoint as the best """
        self._current_best_checkpoint_idx = epoch_idx

    @property
    def current_best_checkpoint_idx(self) -> typing.Union[int, None]:
        return self._current_best_checkpoint_idx


def create(checkpoint_frequency=0, metric=None, metric_mode='min', store_best=False):
    return ClassicCheckpointStrategy(
        checkpoint_frequency=checkpoint_frequency,
        metric=metric,
        metric_mode=metric_mode,
        store_best=store_best
    )
